# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dataclasses
from typing import Any, Optional

from oumi.core.configs.params.base_params import BaseParams


@dataclasses.dataclass
class GuidedDecodingParams(BaseParams):
    """Parameters for guided decoding.

    The parameters are mutually exclusive. Only one of the parameters can be
    specified at a time.
    """

    # Should be Union[dict, BaseModel, str], but omegaconf does not like Union
    json: Optional[Any] = None
    """JSON schema, Pydantic model, or string to guide the output format.

    Can be a dict containing a JSON schema, a Pydantic model class, or a string
    containing JSON schema. Used to enforce structured output from the model.
    """

    regex: Optional[str] = None
    """Regular expression pattern to guide the output format.

    Pattern that the model output must match. Can be used to enforce specific
    text formats or patterns.
    """

    choice: Optional[list[str]] = None
    """List of allowed choices for the output.

    Restricts model output to one of the provided choices. Useful for forcing
    the model to select from a predefined set of options.
    """

    def __post_init__(self) -> None:
        """Validate parameters."""
        provided = sum(x is not None for x in [self.json, self.regex, self.choice])
        if provided > 1:
            raise ValueError(
                "Only one of 'json', 'regex', or 'choice' can be specified"
            )



from unittest.mock import MagicMock, patch

import pytest

from oumi.core.trainers.verl_grpo_trainer import VerlGrpoTrainer

try:
    verl_import_failed = False
    import verl  # pyright: ignore[reportMissingImports]  # noqa: F401
except ModuleNotFoundError:
    verl_import_failed = True


@pytest.mark.skipif(verl_import_failed, reason="verl not available")
def test_init_without_verl():
    with patch("oumi.core.trainers.verl_grpo_trainer.verl", None):
        with pytest.raises(RuntimeError, match="verl is not installed"):
            VerlGrpoTrainer(
                processing_class=MagicMock(),
                config=MagicMock(),
                reward_funcs=[MagicMock()],
                train_dataset=MagicMock(),
                eval_dataset=MagicMock(),
            )


@pytest.mark.skipif(verl_import_failed, reason="verl not available")
def test_init_with_multiple_reward_funcs():
    with pytest.raises(ValueError, match="We only support up to one reward function"):
        VerlGrpoTrainer(
            processing_class=MagicMock(),
            config=MagicMock(),
            reward_funcs=[MagicMock(), MagicMock()],
            train_dataset=MagicMock(),
            eval_dataset=MagicMock(),
        )



import pytest

from oumi.core.configs.params.remote_params import RemoteParams


def test_remote_params_allows_empty():
    params = RemoteParams()
    params.finalize_and_validate()
    # No exception should be raised


def test_remote_params_validates_backoff_base():
    """Test that retry_backoff_base must be positive."""
    with pytest.raises(ValueError, match="Retry backoff base must be greater than 0"):
        params = RemoteParams(retry_backoff_base=0)
        params.finalize_and_validate()

    with pytest.raises(ValueError, match="Retry backoff base must be greater than 0"):
        params = RemoteParams(retry_backoff_base=-1)
        params.finalize_and_validate()


def test_remote_params_validates_backoff_max():
    """Test that retry_backoff_max is be greater than or equal to retry_backoff_base."""
    with pytest.raises(
        ValueError,
        match="Retry backoff max must be greater than or equal to retry backoff base",
    ):
        params = RemoteParams(retry_backoff_base=2, retry_backoff_max=1)
        params.finalize_and_validate()


def test_remote_params_accepts_valid_backoff():
    """Test that valid backoff parameters are accepted."""
    params = RemoteParams(retry_backoff_base=1, retry_backoff_max=30)
    params.finalize_and_validate()
    # No exception should be raised

    params = RemoteParams(retry_backoff_base=0.5, retry_backoff_max=0.5)
    params.finalize_and_validate()
    # No exception should be raised - equal values are allowed



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass, field
from enum import Enum
from pathlib import Path
from typing import Optional

from oumi.core.configs.params.base_params import BaseParams
from oumi.core.configs.params.peft_params import PeftParams
from oumi.core.configs.params.telemetry_params import TelemetryParams
from oumi.core.configs.params.training_params import (
    TrainerType,
    TrainingParams,
)
from oumi.core.registry import REGISTRY, RegistryType
from oumi.utils.logging import logger
from oumi.utils.str_utils import sanitize_run_name


class TunerType(Enum):
    """Enum representing the supported tuners."""

    OPTUNA = "optuna"


class ParamType(Enum):
    """Enum representing the type of parameter to tune."""

    CATEGORICAL = "categorical"
    INT = "int"
    FLOAT = "float"
    LOGUNIFORM = "loguniform"
    UNIFORM = "uniform"

    @staticmethod
    def verify_param_spec(
        param_name: str,
        param_spec: dict,
        valid_training_params: set[str],
        type_name: str = "",
    ) -> None:
        """Verifies that a parameter specification is valid."""
        if param_name not in valid_training_params:
            raise ValueError(
                f"Invalid tunable parameter: {param_name}. "
                f"Must be a valid `{type_name}` field."
            )
        elif isinstance(param_spec, dict):
            # Validate required keys
            if "type" not in param_spec:
                raise ValueError(
                    f"Tunable parameter '{param_name}' must have 'type' key"
                )

            param_type_str = param_spec["type"]

            # Validate type is a valid ParamType
            try:
                param_type = ParamType(param_type_str)
            except ValueError:
                valid_types = [t.value for t in ParamType]
                raise ValueError(
                    f"Invalid type '{param_type_str}' for parameter"
                    f" '{param_name}'. Must be one of: {valid_types}"
                )
            # Validate based on parameter type
            if param_type == ParamType.CATEGORICAL:
                if "choices" not in param_spec:
                    raise ValueError(
                        f"Categorical parameter '{param_name}' must have 'choices' key"
                    )
                if (
                    not isinstance(param_spec["choices"], list)
                    or len(param_spec["choices"]) == 0
                ):
                    raise ValueError(
                        f"Categorical parameter '{param_name}' must have"
                        " non-empty choices list"
                    )
            else:
                # All other types need low and high
                required_keys = {"low", "high"}
                if not required_keys.issubset(param_spec.keys()):
                    raise ValueError(
                        f"Parameter '{param_name}' must have 'low' and 'high' keys"
                    )
        else:
            raise ValueError(f"Tunable parameter '{param_name}' must be a dict")


@dataclass
class TuningParams(BaseParams):
    n_trials: int = field(default_factory=int)
    """Number of tuning trials to perform.

    This defines how many different hyperparameter configurations will be evaluated
    during the tuning process.
    """

    tunable_training_params: dict[str, dict] = field(default_factory=dict)
    """Dictionary mapping parameter names to their search spaces.

    Each value should be a dict specifying the parameter type and range::

        {
            "type": "float",  # or ParamType enum value
            "low": 1e-5,
            "high": 1e-2
        }

    Supported types from ParamType enum:

    - CATEGORICAL: List of discrete choices
    - INT: Integer range
    - FLOAT: Float range (uniform sampling)
    - LOGUNIFORM: Float range with log-scale sampling
    - UNIFORM: Float range with uniform sampling
    """

    fixed_training_params: dict = field(default_factory=dict)
    """A dictionary containing the fixed parameters for the training process.

    These parameters will remain constant throughout the tuning process and will not be
    modified. This allows you to set certain training parameters that you do not wish to
    tune, while still allowing other parameters to be optimized.
    """

    tunable_peft_params: dict[str, dict] = field(default_factory=dict)
    """Dictionary mapping PEFT parameter names to their search spaces.

    Each value should be a dict specifying the parameter type and range::

        {
            "type": "categorical",  # or ParamType enum value
            "choices": ["8", "16", "32"]
        }

    Supported types from ParamType enum:

    - CATEGORICAL: List of discrete choices
    - INT: Integer range
    - FLOAT: Float range (uniform sampling)
    - LOGUNIFORM: Float range with log-scale sampling
    - UNIFORM: Float range with uniform sampling
    """

    fixed_peft_params: dict = field(default_factory=dict)
    """A dictionary containing the fixed parameters for the PEFT configuration.

    These parameters will remain constant throughout the tuning process and will not be
    modified. This allows you to set certain PEFT parameters that you do not wish to
    tune, while still allowing other parameters to be optimized.
    """

    output_dir: str = "output"
    """Directory where the output files will be saved.

    This includes all trained models, evaluation results, and any other artifacts
    produced during the tuning process.
    """

    tuning_study_name: Optional[str] = "oumi-tuning"
    """A unique identifier for the current tuning run.

    This name is used to identify the tuning study in logging outputs, saved model
    checkpoints, and experiment tracking tools like Weights & Biases or
    TensorBoard.
    """

    evaluation_metrics: list[str] = field(default_factory=lambda: ["eval_loss"])
    """The metrics used to evaluate the performance of the model during tuning.

    These metrics are used to compare different hyperparameter configurations and select
    the best-performing parameter configuration.
    """

    evaluation_direction: list[str] = field(default_factory=lambda: ["minimize"])
    """The direction of optimization for each evaluation metric.

    This can be either "minimize" or "maximize", depending on whether lower or higher
    values of the evaluation metric indicate better performance. If only one value is
    provided, it is applied to all metrics.
    """

    log_level: str = "info"
    """The logging level for the main Oumi logger.

    Possible values are "debug", "info", "warning", "error", "critical".
    """

    logging_strategy: str = "trials"
    """The strategy to use for logging during the tuning process.

    Possible values are:
    - "trials": Log every `logging_steps` steps.
    - "epoch": Log at the end of each epoch for each trail configuration.
    - "no": Disable logging.
    """

    logging_dir: Optional[str] = None
    """The directory where training logs will be saved.

    If not specified, defaults to a "logs" subdirectory within the `output_dir`.
    """

    log_examples: bool = False
    """Whether to log an example of the data in the first step for debugging purposes.

    If True, the example will be logged to the console.
    """

    telemetry: TelemetryParams = field(default_factory=TelemetryParams)
    """Parameters for telemetry.

    This field contains telemetry configuration options.
    """

    load_if_exists: bool = field(default_factory=bool)
    """Whether to load an existing tuning study if it exists.

    If True, the tuner will attempt to load a previously saved tuning study from disk.
    If no existing study is found, a new one will be created.
    """

    storage: Optional[str] = None
    """The storage URL for the tuning study.

    This can be a database URL (e.g., SQLite, PostgreSQL) or a file path for local
    storage. If not specified, the study will be stored in memory and will not persist.

    NOTE: In-memory storage does not support loading existing studies later.
    If something breaks during tuning, all progress will be lost.
    """

    tuner_type: TunerType = TunerType.OPTUNA
    """The type of tuner to use for hyperparameter optimization.

    Possible values are:
        - OPTUNA: Optuna tuner.
    """

    trainer_type: TrainerType = TrainerType.TRL_SFT
    """The type of trainer to use for model training.

    Possible values are:

    - TRL_SFT: TRL's SFT Trainer

    TODO: Add more options in the future.
    """

    tuner_sampler: Optional[str] = None
    """The sampler to use for the tuner.

    This is specific to the tuner type. For Optuna, this could be "TPESampler",
    "RandomSampler", etc. If not specified, the default sampler for the tuner will be
    used.
    """

    custom_eval_metrics: Optional[list[str]] = field(default_factory=list)
    """Custom evaluation metrics.

    This specifies if the tuner will use user defined evaluation metrics to tune the
    model parameters.
    """

    def __post_init__(self):
        """Verifies params."""
        if self.logging_dir is None:
            self.logging_dir = f"{self.output_dir}/logs"

        # Validate logging strategy
        valid_logging_strategies = {"trials", "epoch", "no"}
        if self.logging_strategy not in valid_logging_strategies:
            raise ValueError(
                f"Invalid logging_strategy: {self.logging_strategy}. "
                f"Choose from {valid_logging_strategies}."
            )

        # Validate evaluation metrics and directions
        if len(self.evaluation_metrics) != len(self.evaluation_direction):
            if len(self.evaluation_direction) == 1:
                self.evaluation_direction *= len(self.evaluation_metrics)
                logger.warning(
                    "Single evaluation_direction provided. "
                    "Applying it to all evaluation_metrics."
                )
            else:
                raise ValueError(
                    "Length of evaluation_metrics must match length of "
                    "evaluation_direction, or evaluation_direction must be of length 1."
                )

        # Validate each evaluation direction
        for direction in self.evaluation_direction:
            if direction not in {"minimize", "maximize"}:
                raise ValueError(
                    f"Invalid evaluation_direction: {direction}. "
                    'Choose either "minimize" or "maximize".'
                )

        # Warn if using multiple metrics with incompatible logging strategy
        if len(self.evaluation_metrics) > 1 and self.logging_strategy == "epoch":
            logger.warning(
                "Using multiple evaluation_metrics with 'epoch' logging_strategy "
                "may lead to ambiguous logging. Consider using 'trials' instead."
            )

        # Validate trainer type
        # TODO: Add more options in the future.
        if self.trainer_type != TrainerType.TRL_SFT:
            raise ValueError(
                f"Invalid trainer_type: {self.trainer_type}. "
                f"Choose from {[t.value for t in [TrainerType.TRL_SFT]]}."
            )

        # Validate that the params keys are valid TrainingParams fields
        valid_training_params = {
            field.name for field in TrainingParams.__dataclass_fields__.values()
        }

        # Verify fixed training params keys are valid TrainingParams fields
        for param_name in self.fixed_training_params.keys():
            if param_name not in valid_training_params:
                raise ValueError(
                    f"Invalid fixed parameter: {param_name}. "
                    f"Must be a valid `TrainingParams` field."
                )

        # Ensure tunable_training_params values are valid
        for param_name, param_spec in self.tunable_training_params.items():
            ParamType.verify_param_spec(
                param_name, param_spec, valid_training_params, "TrainingParams"
            )

        # Validate that the params keys are valid PEFTParams fields
        valid_training_params = {
            field.name for field in PeftParams.__dataclass_fields__.values()
        }
        # Verify fixed training params keys are valid PEFT fields
        for param_name in self.fixed_peft_params.keys():
            if param_name not in valid_training_params:
                raise ValueError(
                    f"Invalid fixed parameter: {param_name}. "
                    f"Must be a valid `PeftParams` field."
                )

        # Ensure tunable_training_params values are valid
        for param_name, param_spec in self.tunable_peft_params.items():
            ParamType.verify_param_spec(
                param_name, param_spec, valid_training_params, "PeftParams"
            )

        self.tuning_study_name = sanitize_run_name(self.tuning_study_name)

        # Validate custom evaluation metrics are registered in Oumi
        if self.custom_eval_metrics:
            try:
                import oumi.evaluation.registry  # noqa: F401
            except Exception:
                # Best-effort: continue, REGISTRY decorator may still lazy-load
                pass

            unknown: list[str] = []
            for name in self.custom_eval_metrics:
                if not isinstance(name, str) or not name:
                    unknown.append(str(name))
                    continue
                if REGISTRY.get_evaluation_function(name) is None:
                    unknown.append(name)

            if unknown:
                available = sorted(
                    REGISTRY.get_all(RegistryType.EVALUATION_FUNCTION).keys()
                )
                raise ValueError(
                    "Unregistered custom_eval_metrics detected: "
                    f"{unknown}. Available evaluation functions: {available}"
                )

    @property
    def telemetry_dir(self) -> Optional[Path]:
        """Returns the telemetry stats output directory."""
        result: Optional[Path] = None
        if self.telemetry.telemetry_dir:
            result = Path(self.telemetry.telemetry_dir)

        if self.output_dir:
            output_dir = Path(self.output_dir)
            # If `telemetry.telemetry_dir` is relative, then treat it
            # as a sub-directory of `output_dir`.
            if result and not result.is_absolute():
                result = output_dir / result

        return result



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
from typing import Any, Optional

from typing_extensions import override

from oumi.core.configs import GenerationParams, ModelParams, RemoteParams
from oumi.core.types.conversation import Conversation
from oumi.inference.remote_inference_engine import RemoteInferenceEngine


class OpenAIInferenceEngine(RemoteInferenceEngine):
    """Engine for running inference against the OpenAI API."""

    @property
    @override
    def base_url(self) -> Optional[str]:
        """Return the default base URL for the OpenAI API."""
        return "https://api.openai.com/v1/chat/completions"

    @property
    @override
    def api_key_env_varname(self) -> Optional[str]:
        """Return the default environment variable name for the OpenAI API key."""
        return "OPENAI_API_KEY"

    @override
    def _convert_conversation_to_api_input(
        self,
        conversation: Conversation,
        generation_params: GenerationParams,
        model_params: ModelParams,
    ) -> dict[str, Any]:
        """Converts a conversation to an OpenAI input.

        Documentation: https://platform.openai.com/docs/api-reference/chat/create

        Args:
            conversation: The conversation to convert.
            generation_params: Parameters for generation during inference.
            model_params: Model parameters to use during inference.

        Returns:
            Dict[str, Any]: A dictionary representing the OpenAI input.
        """
        if model_params.model_name == "o1-preview":
            generation_params = copy.deepcopy(generation_params)

            # o1-preview does NOT support logit_bias.
            generation_params.logit_bias = {}

            # o1-preview only supports temperature = 1.
            generation_params.temperature = 1.0

        return super()._convert_conversation_to_api_input(
            conversation=conversation,
            generation_params=generation_params,
            model_params=model_params,
        )

    @override
    def _default_remote_params(self) -> RemoteParams:
        """Returns the default remote parameters."""
        return RemoteParams(num_workers=50, politeness_policy=60.0)



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Clouds module for the Oumi (Open Universal Machine Intelligence) library.

This module provides implementations for various cloud platforms that can be used
with the Oumi launcher for running and managing jobs.

Example:
    >>> from oumi.launcher import JobConfig, JobResources
    >>> from oumi.launcher.clouds import LocalCloud
    >>> local_cloud = LocalCloud()
    >>> job_resources = JobResources(cloud="local")
    >>> job_config = JobConfig(
    ...     name="my_job", resources=job_resources, run="python train.py"
    ... )
    >>> job_status = local_cloud.up_cluster(job_config, name="my_cluster")

Note:
    Ensure that you have the necessary credentials and configurations set up
    for the cloud platform you intend to use.
"""

from oumi.launcher.clouds.frontier_cloud import FrontierCloud
from oumi.launcher.clouds.local_cloud import LocalCloud
from oumi.launcher.clouds.perlmutter_cloud import PerlmutterCloud
from oumi.launcher.clouds.polaris_cloud import PolarisCloud
from oumi.launcher.clouds.sky_cloud import SkyCloud
from oumi.launcher.clouds.slurm_cloud import SlurmCloud
from oumi.utils import logging

logging.configure_dependency_warnings()


__all__ = [
    "FrontierCloud",
    "LocalCloud",
    "PerlmutterCloud",
    "PolarisCloud",
    "SkyCloud",
    "SlurmCloud",
]



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Sequence
from pprint import pformat
from typing import NamedTuple, Optional

from oumi.utils.logging import logger

try:
    # The library is only useful for NVIDIA GPUs, and
    # may not be installed for other vendors e.g., AMD
    import pynvml  # pyright: ignore[reportMissingImports]
except ModuleNotFoundError:
    pynvml = None

# TODO: OPE-562 - Add support for `amdsmi.amdsmi_init()`` for AMD GPUs


def _initialize_pynvml() -> bool:
    """Attempts to initialize pynvml library. Returns True on success."""
    global pynvml
    if pynvml is None:
        return False

    try:
        pynvml.nvmlInit()
    except Exception:
        logger.error(
            "Failed to initialize pynvml library. All pynvml calls will be disabled."
        )
        pynvml = None

    return pynvml is not None


def _initialize_pynvml_and_get_pynvml_device_count() -> Optional[int]:
    """Attempts to initialize pynvml library.

    Returns device count on success, or None otherwise.
    """
    global pynvml
    # The call to `pynvml is None` is technically redundant but exists here
    # to make pyright happy.
    if pynvml is None or not _initialize_pynvml():
        return None
    return int(pynvml.nvmlDeviceGetCount())


class NVidiaGpuRuntimeInfo(NamedTuple):
    """Contains misc NVIDIA GPU measurements and stats retrieved by `pynvml`.

    The majority of fields are optional. You can control whether they are
    populated by setting boolean query parameters of
    `_get_nvidia_gpu_runtime_info_impl(, ...)` such as `memory`, `temperature`,
    `fan_speed`, etc.
    """

    device_index: int
    """Zero-based device index."""

    device_count: int
    """Total number of GPU devices on this node."""

    used_memory_mb: Optional[float] = None
    """Used GPU memory in MB."""

    temperature: Optional[int] = None
    """GPU temperature in Celcius."""

    fan_speed: Optional[int] = None
    """GPU fan speed in [0,100] range."""

    fan_speeds: Optional[Sequence[int]] = None
    """An array of GPU fan speeds.

    The array's length is equal to the number of fans per GPU (can be multiple).
    Speed values are in [0, 100] range.
    """

    power_usage_watts: Optional[float] = None
    """GPU power usage in Watts."""

    power_limit_watts: Optional[float] = None
    """GPU power limit in Watts."""

    gpu_utilization: Optional[int] = None
    """GPU compute utilization. Range: [0,100]."""

    memory_utilization: Optional[int] = None
    """GPU memory utilization. Range: [0,100]."""

    performance_state: Optional[int] = None
    """See `nvmlPstates_t`. Valid values are in [0,15] range, or 32 if unknown.

    0 corresponds to Maximum Performance.
    15 corresponds to Minimum Performance.
    """

    clock_speed_graphics: Optional[int] = None
    """Graphics clock speed (`NVML_CLOCK_GRAPHICS`) in MHz."""

    clock_speed_sm: Optional[int] = None
    """SM clock speed (`NVML_CLOCK_SM`) in MHz."""

    clock_speed_memory: Optional[int] = None
    """Memory clock speed (`NVML_CLOCK_MEM`) in MHz."""


def _get_nvidia_gpu_runtime_info_impl(
    device_index: int = 0,
    *,
    memory: bool = False,
    temperature: bool = False,
    fan_speed: bool = False,
    power_usage: bool = False,
    utilization: bool = False,
    performance_state: bool = False,
    clock_speed: bool = False,
) -> Optional[NVidiaGpuRuntimeInfo]:
    global pynvml
    if pynvml is None:
        return None

    device_count = _initialize_pynvml_and_get_pynvml_device_count()
    if device_count is None or device_count <= 0:
        return None
    elif device_index < 0 or device_index >= device_count:
        raise ValueError(
            f"Device index ({device_index}) must be "
            f"within the [0, {device_count}) range."
        )

    try:
        gpu_handle = pynvml.nvmlDeviceGetHandleByIndex(device_index)
    except pynvml.NVMLError_NotSupported:  # pyright: ignore
        # This error is expected on some systems.
        # Only do DEBUG-level logging to reduce noise.
        logger.debug(f"pyNVML GPU handle not supported for device: {device_index}")
    except Exception:
        logger.exception(f"Failed to get GPU handle for device: {device_index}")
        return None

    used_memory_mb_value: Optional[float] = None
    if memory:
        try:
            info = pynvml.nvmlDeviceGetMemoryInfo(gpu_handle)
            used_memory_mb_value = float(info.used) // 1024**2
        except pynvml.NVMLError_NotSupported:  # pyright: ignore
            # This error is expected on some systems.
            # Only do DEBUG-level logging to reduce noise.
            logger.debug(
                f"pyNVML GPU memory info not supported for device: {device_index}"
            )
        except Exception:
            logger.exception(
                f"Failed to get GPU memory info for device: {device_index}"
            )
            return None

    temperature_value: Optional[int] = None
    if temperature:
        try:
            temperature_value = pynvml.nvmlDeviceGetTemperature(
                gpu_handle, pynvml.NVML_TEMPERATURE_GPU
            )
        except pynvml.NVMLError_NotSupported:  # pyright: ignore
            # This error is expected on some systems.
            # Only do DEBUG-level logging to reduce noise.
            logger.debug(
                f"pyNVML GPU temperature not supported for device: {device_index}"
            )
        except Exception:
            logger.exception(
                f"Failed to get GPU temperature for device: {device_index}"
            )
            return None

    fan_speed_value: Optional[int] = None
    fan_speeds_value: Optional[Sequence[int]] = None
    if fan_speed:
        try:
            fan_speed_value = pynvml.nvmlDeviceGetFanSpeed(gpu_handle)
        except pynvml.NVMLError_NotSupported:  # pyright: ignore
            # This error is expected on some systems.
            # Only do DEBUG-level logging to reduce noise.
            logger.debug(
                f"pyNVML GPU fan speed not supported for device: {device_index}"
            )
        except Exception:
            # The `GetFanSpeed` function fails on many systems
            # Only do DEBUG-level logging to reduce noise.
            logger.debug(
                f"Failed to get GPU fan speed for device: {device_index}", exc_info=True
            )

        if fan_speed_value is not None:
            fan_speeds_value = tuple([fan_speed_value])
            if hasattr(pynvml, "nvmlDeviceGetNumFans"):
                try:
                    fan_count = pynvml.nvmlDeviceGetNumFans(gpu_handle)
                    value = [0] * fan_count
                    for i in range(fan_count):
                        speed = pynvml.nvmlDeviceGetFanSpeed_v2(gpu_handle, i)
                        value[i] = speed
                    # Make it immutable.
                    fan_speeds_value = tuple(value)
                except Exception:
                    fan_speeds_value = tuple([fan_speed_value])

    power_usage_watts_value: Optional[float] = None
    power_limit_watts_value: Optional[float] = None
    if power_usage:
        try:
            milliwatts = pynvml.nvmlDeviceGetPowerUsage(gpu_handle)
            power_usage_watts_value = float(milliwatts) * 1e-3

            milliwatts = pynvml.nvmlDeviceGetPowerManagementLimit(gpu_handle)
            power_limit_watts_value = float(milliwatts) * 1e-3
        except pynvml.NVMLError_NotSupported:  # pyright: ignore
            # This error is expected on some systems.
            # Only do DEBUG-level logging to reduce noise.
            logger.debug(
                f"pyNVML GPU power usage not supported for device: {device_index}"
            )
        except Exception:
            logger.exception(
                f"Failed to get GPU power usage for device: {device_index}"
            )
            return None

    gpu_utilization_value: Optional[float] = None
    memory_utilization_value: Optional[float] = None
    if utilization:
        try:
            result = pynvml.nvmlDeviceGetUtilizationRates(gpu_handle)
            gpu_utilization_value = int(result.gpu)
            memory_utilization_value = int(result.memory)
        except pynvml.NVMLError_NotSupported:  # pyright: ignore
            # This error is expected on some systems.
            # Only do DEBUG-level logging to reduce noise.
            logger.debug(
                f"pyNVML GPU utilization not supported for device: {device_index}"
            )
        except Exception:
            logger.exception(
                f"Failed to get GPU utilization for device: {device_index}"
            )
            return None

    performance_state_value: Optional[int] = None
    if performance_state:
        try:
            performance_state_value = int(
                pynvml.nvmlDeviceGetPerformanceState(gpu_handle)
            )
        except pynvml.NVMLError_NotSupported:  # pyright: ignore
            # This error is expected on some systems.
            # Only do DEBUG-level logging to reduce noise.
            logger.debug(
                f"pyNVML GPU performance state not supported for device: {device_index}"
            )
        except Exception:
            logger.exception(
                f"Failed to get GPU performance state for device: {device_index}"
            )
            return None

    clock_speed_graphics_value: Optional[int] = None
    clock_speed_sm_value: Optional[int] = None
    clock_speed_memory_value: Optional[int] = None
    if clock_speed:
        try:
            clock_speed_graphics_value = int(
                pynvml.nvmlDeviceGetClockInfo(gpu_handle, pynvml.NVML_CLOCK_GRAPHICS)
            )
            clock_speed_sm_value = int(
                pynvml.nvmlDeviceGetClockInfo(gpu_handle, pynvml.NVML_CLOCK_SM)
            )
            clock_speed_memory_value = int(
                pynvml.nvmlDeviceGetClockInfo(gpu_handle, pynvml.NVML_CLOCK_MEM)
            )
        except pynvml.NVMLError_NotSupported:  # pyright: ignore
            # This error is expected on some systems.
            # Only do DEBUG-level logging to reduce noise.
            logger.debug(
                f"pyNVML GPU clock speed not supported for device: {device_index}"
            )
        except Exception:
            logger.exception(
                f"Failed to get GPU clock speed for device: {device_index}"
            )
            return None

    return NVidiaGpuRuntimeInfo(
        device_index=device_index,
        device_count=device_count,
        used_memory_mb=used_memory_mb_value,
        temperature=temperature_value,
        fan_speed=fan_speed_value,
        fan_speeds=fan_speeds_value,
        power_usage_watts=power_usage_watts_value,
        power_limit_watts=power_limit_watts_value,
        gpu_utilization=gpu_utilization_value,
        memory_utilization=memory_utilization_value,
        performance_state=performance_state_value,
        clock_speed_graphics=clock_speed_graphics_value,
        clock_speed_sm=clock_speed_sm_value,
        clock_speed_memory=clock_speed_memory_value,
    )


def get_nvidia_gpu_runtime_info(
    device_index: int = 0,
) -> Optional[NVidiaGpuRuntimeInfo]:
    """Returns runtime stats for Nvidia GPU."""
    return _get_nvidia_gpu_runtime_info_impl(
        device_index=device_index,
        memory=True,
        temperature=True,
        fan_speed=True,
        power_usage=True,
        utilization=True,
        performance_state=True,
        clock_speed=True,
    )


def log_nvidia_gpu_runtime_info(device_index: int = 0, log_prefix: str = "") -> None:
    """Prints the current NVIDIA GPU runtime info."""
    info = get_nvidia_gpu_runtime_info(device_index)
    logger.info(f"{log_prefix.rstrip()} GPU runtime info: {pformat(info)}.")


def get_nvidia_gpu_memory_utilization(device_index: int = 0) -> float:
    """Returns amount of memory being used on an Nvidia GPU in MiB."""
    info: Optional[NVidiaGpuRuntimeInfo] = _get_nvidia_gpu_runtime_info_impl(
        device_index=device_index, memory=True
    )
    return (
        info.used_memory_mb
        if (info is not None and info.used_memory_mb is not None)
        else 0.0
    )


def log_nvidia_gpu_memory_utilization(
    device_index: int = 0, log_prefix: str = ""
) -> None:
    """Prints amount of memory being used on an Nvidia GPU."""
    memory_mib = get_nvidia_gpu_memory_utilization(device_index)
    logger.info(f"{log_prefix.rstrip()} GPU memory occupied: {memory_mib} MiB.")


def get_nvidia_gpu_temperature(device_index: int = 0) -> int:
    """Returns the current temperature readings for the device, in degrees C."""
    info: Optional[NVidiaGpuRuntimeInfo] = _get_nvidia_gpu_runtime_info_impl(
        device_index=device_index,
        temperature=True,
    )
    return (
        info.temperature if (info is not None and info.temperature is not None) else 0
    )


def log_nvidia_gpu_temperature(device_index: int = 0, log_prefix: str = "") -> None:
    """Prints the current temperature readings for the device, in degrees C."""
    temperature = get_nvidia_gpu_temperature(device_index)
    logger.info(f"{log_prefix.rstrip()} GPU temperature: {temperature} C.")


def get_nvidia_gpu_fan_speeds(device_index: int = 0) -> Sequence[int]:
    """Returns the current fan speeds for NVIDIA GPU device."""
    info: Optional[NVidiaGpuRuntimeInfo] = _get_nvidia_gpu_runtime_info_impl(
        device_index=device_index, fan_speed=True
    )
    return (
        info.fan_speeds
        if (info is not None and info.fan_speeds is not None)
        else tuple()
    )


def log_nvidia_gpu_fan_speeds(device_index: int = 0, log_prefix: str = "") -> None:
    """Prints the current NVIDIA GPU fan speeds."""
    fan_speeds = get_nvidia_gpu_fan_speeds(device_index)
    logger.info(f"{log_prefix.rstrip()} GPU fan speeds: {fan_speeds}.")


def get_nvidia_gpu_power_usage(device_index: int = 0) -> float:
    """Returns the current power usage for NVIDIA GPU device."""
    info: Optional[NVidiaGpuRuntimeInfo] = _get_nvidia_gpu_runtime_info_impl(
        device_index=device_index, power_usage=True
    )
    return (
        info.power_usage_watts
        if (info is not None and info.power_usage_watts is not None)
        else 0.0
    )


def log_nvidia_gpu_power_usage(device_index: int = 0, log_prefix: str = "") -> None:
    """Prints the current NVIDIA GPU power usage."""
    power_usage = get_nvidia_gpu_power_usage(device_index)
    logger.info(f"{log_prefix.rstrip()} GPU power usage: {power_usage:.2f}W.")



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, Union

import pandas as pd
from typing_extensions import override

from oumi.core.datasets import BaseSftDataset
from oumi.core.registry import register_dataset
from oumi.core.types.conversation import Conversation, Message, Role


@register_dataset("nvidia/ChatRAG-Bench")
class ChatRAGBenchDataset(BaseSftDataset):
    default_dataset: str = "nvidia/ChatRAG-Bench"
    default_system_message: str = (
        "This is a chat between a user and an artificial intelligence assistant. "
        "The assistant gives helpful, detailed, and polite answers to the user's "
        "questions based on the context. The assistant should also indicate when "
        "the answer cannot be found in the context."
    )
    default_subset: str = "doc2dial"

    def __init__(
        self,
        *,
        split: str = "test",
        task: str = "generation",
        subset: Optional[str] = None,
        num_context_docs: int = 5,
        **kwargs,
    ) -> None:
        """Initialize the ChatRag dataset.

        Args:
            split: The split of the dataset to use. Defaults to "test".
            num_context_docs: The number of context documents to include in each
            example.
            subset: The subset of the dataset to use. Defaults to None.
            task: The task for which the dataset is intended. Defaults to "generation".
            **kwargs: Additional keyword arguments to be passed to the base class.
        """
        self.num_context_docs = num_context_docs

        subset = subset or self.default_subset

        # This dataset is for evaluation only and does not contain a training split.
        if split != "test":
            raise ValueError(
                f"This dataset only supports the 'test' split. Got: {split}"
            )

        if task != "generation":
            raise ValueError("This dataset can only be used for evaluation tasks")

        # Get the test split name for this subset, which may be different
        # from the Oumi user facing split.
        internal_split = self._get_test_dataset_split(subset)

        super().__init__(
            split=internal_split,
            subset=subset,
            task=task,
            **kwargs,
        )

    def _get_test_dataset_split(self, subset: str) -> str:
        # Most subset use the "test" split, except these three
        # Note: these datasets all have a single split (test or dev)
        # We consider them all to be test datasets.
        if subset in ("coqa", "inscit", "topiocqa"):
            return "dev"
        return "test"

    def _get_instruction(self) -> Optional[str]:
        """Get an appropriate instruction for each dataset subset."""
        subset_instructions = {
            "doc2dial": "Please give a full and complete answer for the question.",
            "quac": "Please give a full and complete answer for the question.",
            "qrecc": "Please give a full and complete answer for the question.",
            "coqa": (
                "Answer the following question with a short span. "
                "The answer needs to be just in a few words."
            ),
            "doqa": "Please give a full and complete answer for the question.",
            "convfinqa": "Please give a full and complete answer for the question.",
            "sqa": "Answer the following question with one or a list of items.",
            "topiocqa": (
                "Answer the following question with a short span, "
                "or a full and complete answer."
            ),
            "hybridial": "Please give a full and complete answer for the question.",
            "inscit": "Please give a full and complete answer for the question.",
        }

        if self.dataset_subset is None:
            raise ValueError("The dataset subset must be specified.")

        return subset_instructions.get(self.dataset_subset)

    def _format_context_document(self, doc: dict) -> str:
        # Not all docs have titles
        if doc["title"] is not None:
            return f"title: {doc['title']}, source: {doc['text']}"
        else:
            return f"source: {doc['text']}"

    @override
    def transform_conversation(self, example: Union[dict, pd.Series]) -> Conversation:
        """Transforms a given example into a Conversation object.

        Args:
            example (Union[dict, pd.Series]): The example to transform.

        Returns:
            Conversation: The transformed Conversation object.
        """
        messages = []

        # Add system message
        messages.append(Message(role=Role.SYSTEM, content=self.default_system_message))

        # Add grounding context
        grounding_documents = example["ctxs"][: self.num_context_docs]
        grounding_context = "\n\n".join(
            [self._format_context_document(doc) for doc in grounding_documents]
        )
        messages.append(Message(role=Role.USER, content="\n\n" + grounding_context))

        # Add conversation history
        added_instruction = False
        for turn in example["messages"]:
            content = turn["content"]

            if turn["role"] == "user" and not added_instruction:
                # Add user instruction to the first user turn
                content = f"{self._get_instruction()} {content}"
                added_instruction = True

            messages.append(Message(role=turn["role"], content=content))

        return Conversation(messages=messages)



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

from oumi.core.configs import JobConfig
from oumi.core.launcher import BaseCloud, BaseCluster, JobStatus
from oumi.core.registry import register_cloud_builder
from oumi.launcher.clients.slurm_client import SlurmClient
from oumi.launcher.clusters.slurm_cluster import SlurmCluster


class SlurmCloud(BaseCloud):
    """A resource pool for managing jobs in Slurm clusters."""

    def __init__(self):
        """Initializes a new instance of the SlurmCloud class."""
        # A mapping from cluster names to Slurm Cluster instances.
        self._clusters = {}

        # Initialize default connections.
        self.initialize_clusters()

    def _get_or_create_cluster(self, name: str) -> SlurmCluster:
        """Gets the cluster with the specified name, or creates one if it doesn't exist.

        Args:
            name: The name of the cluster.

        Returns:
            SlurmCluster: The cluster instance.
        """
        if name not in self._clusters:
            cluster_info = SlurmCluster.parse_cluster_name(name)
            self._clusters[name] = SlurmCluster(
                name,
                SlurmClient(
                    user=cluster_info.user,
                    slurm_host=cluster_info.hostname,
                    cluster_name=cluster_info.name,
                ),
            )
        return self._clusters[name]

    def initialize_clusters(self) -> list[BaseCluster]:
        """Initializes clusters for the specified user for all Slurm queues.

        Returns:
            List[SlurmCluster]: The list of initialized clusters.
        """
        connections = SlurmCluster.get_slurm_connections()
        clusters = []
        for c in connections:
            cluster = self._get_or_create_cluster(c.name)
            clusters.append(cluster)
        return clusters

    def up_cluster(self, job: JobConfig, name: Optional[str], **kwargs) -> JobStatus:
        """Creates a cluster and starts the provided Job."""
        if not job.user:
            raise ValueError("User must be provided in the job config.")
        if name:
            cluster_info = SlurmCluster.parse_cluster_name(name)
            if cluster_info.user != job.user:
                raise ValueError(
                    f"Invalid cluster name: `{name}`. "
                    f"User must match the provided job user: `{job.user}`."
                )
        else:
            raise ValueError(
                "A cluster name must be provided for Slurm. "
                "Cluster names are of the form 'user@hostname'."
            )
        cluster = self._get_or_create_cluster(cluster_info.name)
        job_status = cluster.run_job(job)
        if not job_status:
            raise RuntimeError("Failed to start job.")
        return job_status

    def get_cluster(self, name) -> Optional[BaseCluster]:
        """Gets the cluster with the specified name, or None if not found."""
        clusters = self.list_clusters()
        for cluster in clusters:
            if cluster.name() == name:
                return cluster
        return None

    def list_clusters(self) -> list[BaseCluster]:
        """Lists the active clusters on this cloud."""
        return list(self._clusters.values())


@register_cloud_builder("slurm")
def slurm_cloud_builder() -> SlurmCloud:
    """Builds a SlurmCloud instance."""
    return SlurmCloud()



from multiprocessing import Process, set_start_method
from unittest.mock import Mock, patch

import pytest

from oumi.core.configs import JobConfig, JobResources, StorageMount
from oumi.core.launcher import BaseCloud, BaseCluster, JobState, JobStatus
from oumi.launcher.launcher import (
    LAUNCHER,
    Launcher,
    cancel,
    down,
    get_cloud,
    run,
    status,
    stop,
    up,
    which_clouds,
)


#
# Fixtures
#
@pytest.fixture
def mock_registry():
    with patch("oumi.launcher.launcher.REGISTRY") as registry:
        yield registry


def _get_default_job(cloud: str) -> JobConfig:
    resources = JobResources(
        cloud=cloud,
        region="us-central1",
        zone=None,
        accelerators="A100-80GB",
        cpus="4",
        memory="64",
        instance_type=None,
        use_spot=True,
        disk_size=512,
        disk_tier="low",
    )
    return JobConfig(
        name="myjob",
        user="user",
        working_dir="./",
        num_nodes=2,
        resources=resources,
        envs={"var1": "val1"},
        file_mounts={},
        storage_mounts={
            "~/home/remote/path/gcs/": StorageMount(
                source="gs://mybucket/", store="gcs"
            )
        },
        setup="pip install -r requirements.txt",
        run="./hello_world.sh",
    )


#
# Tests
#
def test_launcher_get_cloud(mock_registry):
    sky_mock = Mock(spec=BaseCloud)
    polaris_mock = Mock(spec=BaseCloud)

    def _sky_builder():
        return sky_mock

    def _polaris_builder():
        return polaris_mock

    mock_registry.get.side_effect = lambda name, reg_type: {
        "sky": _sky_builder,
        "polaris": _polaris_builder,
    }.get(name)
    launcher = Launcher()
    cloud = launcher.get_cloud(_get_default_job("sky"))
    assert cloud == sky_mock
    assert cloud != polaris_mock


def test_launcher_get_cloud_missing_value(mock_registry):
    with pytest.raises(ValueError) as exception_info:
        sky_mock = Mock(spec=BaseCloud)
        polaris_mock = Mock(spec=BaseCloud)

        def _sky_builder():
            return sky_mock

        def _polaris_builder():
            return polaris_mock

        mock_registry.get.return_value = None
        mock_registry.get_all.return_value = {
            "sky": _sky_builder,
            "polaris": _polaris_builder,
        }
        launcher = Launcher()
        launcher.get_cloud(_get_default_job("lambda"))
    assert "not found in the registry." in str(exception_info.value)


def test_launcher_get_cloud_empty(mock_registry):
    with pytest.raises(ValueError) as exception_info:
        mock_registry.get_all.return_value = {}
        mock_registry.get.return_value = None
        launcher = Launcher()
        launcher.get_cloud(_get_default_job("sky"))
    assert "not found in the registry." in str(exception_info.value)


def test_launcher_get_cloud_by_name_missing_value(mock_registry):
    with pytest.raises(ValueError) as exception_info:
        sky_mock = Mock(spec=BaseCloud)
        polaris_mock = Mock(spec=BaseCloud)

        def _sky_builder():
            return sky_mock

        def _polaris_builder():
            return polaris_mock

        mock_registry.get_all.return_value = {
            "sky": _sky_builder,
            "polaris": _polaris_builder,
        }
        mock_registry.get.return_value = None
        launcher = Launcher()
        _ = launcher.get_cloud("lambda")
    assert "not found in the registry." in str(exception_info.value)


def test_launcher_get_cloud_by_name_empty(mock_registry):
    with pytest.raises(ValueError) as exception_info:
        mock_registry.get_all.return_value = {}
        mock_registry.get.return_value = None
        launcher = Launcher()
        _ = launcher.get_cloud("lambda")
    assert "not found in the registry." in str(exception_info.value)


def test_launcher_up_succeeds(mock_registry):
    mock_cluster = Mock(spec=BaseCluster)
    mock_cloud = Mock(spec=BaseCloud)

    def _builder():
        return mock_cloud

    mock_registry.get_all.return_value = {
        "custom": _builder,
    }
    mock_registry.get.return_value = _builder
    expected_job_status = JobStatus(
        id="job_id",
        cluster="custom",
        name="foo",
        status="running",
        metadata="bar",
        done=False,
        state=JobState.PENDING,
    )
    mock_cloud.up_cluster.return_value = expected_job_status
    mock_cloud.get_cluster.return_value = mock_cluster
    launcher = Launcher()
    job = _get_default_job("custom")
    result = launcher.up(job, "custom")
    mock_cloud.up_cluster.assert_called_once_with(job, "custom")
    mock_cloud.get_cluster.assert_called_once_with("custom")
    assert result == (mock_cluster, expected_job_status)


def test_launcher_up_succeeds_kwargs(mock_registry):
    mock_cluster = Mock(spec=BaseCluster)
    mock_cloud = Mock(spec=BaseCloud)

    def _builder():
        return mock_cloud

    mock_registry.get_all.return_value = {
        "custom": _builder,
    }
    mock_registry.get.return_value = _builder
    expected_job_status = JobStatus(
        id="job_id",
        cluster="custom",
        name="foo",
        status="running",
        metadata="bar",
        done=False,
        state=JobState.PENDING,
    )
    mock_cloud.up_cluster.return_value = expected_job_status
    mock_cloud.get_cluster.return_value = mock_cluster
    launcher = Launcher()
    job = _get_default_job("custom")
    result = launcher.up(job, "custom", foo="bar")
    mock_cloud.up_cluster.assert_called_once_with(job, "custom", foo="bar")
    mock_cloud.get_cluster.assert_called_once_with("custom")
    assert result == (mock_cluster, expected_job_status)


def test_launcher_up_succeeds_no_name(mock_registry):
    mock_cluster = Mock(spec=BaseCluster)
    mock_cloud = Mock(spec=BaseCloud)

    def _builder():
        return mock_cloud

    mock_registry.get_all.return_value = {
        "custom": _builder,
    }
    mock_registry.get.return_value = _builder
    expected_job_status = JobStatus(
        id="job_id",
        cluster="custom",
        name="foo",
        status="running",
        metadata="bar",
        done=False,
        state=JobState.PENDING,
    )
    mock_cloud.up_cluster.return_value = expected_job_status
    mock_cloud.get_cluster.return_value = mock_cluster
    launcher = Launcher()
    job = _get_default_job("custom")
    result = launcher.up(job, None)
    mock_cloud.up_cluster.assert_called_once_with(job, None)
    mock_cloud.get_cluster.assert_called_once_with("custom")
    assert result == (mock_cluster, expected_job_status)


def test_launcher_up_inavlid_cluster(mock_registry):
    with pytest.raises(RuntimeError) as exception_info:
        mock_cloud = Mock(spec=BaseCloud)

        def _builder():
            return mock_cloud

        mock_registry.get_all.return_value = {
            "custom": _builder,
        }
        mock_registry.get.return_value = _builder
        expected_job_status = JobStatus(
            id="job_id",
            cluster="custom",
            name="foo",
            status="running",
            metadata="bar",
            done=False,
            state=JobState.PENDING,
        )
        mock_cloud.up_cluster.return_value = expected_job_status
        mock_cloud.get_cluster.return_value = None
        launcher = Launcher()
        job = _get_default_job("custom")
        launcher.up(job, None)
    assert "not found" in str(exception_info.value)


def test_launcher_run_succeeds(mock_registry):
    mock_cluster = Mock(spec=BaseCluster)
    mock_cloud = Mock(spec=BaseCloud)

    def _builder():
        return mock_cloud

    mock_registry.get_all.return_value = {
        "custom": _builder,
    }
    mock_registry.get.return_value = _builder
    expected_job_status = JobStatus(
        id="job_id",
        cluster="custom",
        name="foo",
        status="running",
        metadata="bar",
        done=False,
        state=JobState.PENDING,
    )
    mock_cloud.get_cluster.return_value = mock_cluster
    mock_cluster.run_job.return_value = expected_job_status
    launcher = Launcher()
    job = _get_default_job("custom")
    result = launcher.run(job, "custom")
    mock_cloud.get_cluster.assert_called_once_with("custom")
    mock_cluster.run_job.assert_called_once_with(job)
    assert result == expected_job_status


def test_launcher_run_fails(mock_registry):
    with pytest.raises(ValueError) as exception_info:
        mock_cloud = Mock(spec=BaseCloud)

        def _builder():
            return mock_cloud

        mock_registry.get_all.return_value = {
            "custom": _builder,
        }
        mock_registry.get.return_value = _builder
        mock_cloud.get_cluster.return_value = None
        launcher = Launcher()
        job = _get_default_job("custom")
        launcher.run(job, "custom")
    assert "not found" in str(exception_info.value)


def test_launcher_cancel_succeeds(mock_registry):
    mock_cluster = Mock(spec=BaseCluster)
    mock_cloud = Mock(spec=BaseCloud)

    def _builder():
        return mock_cloud

    mock_registry.get_all.return_value = {
        "cloud": _builder,
    }
    mock_registry.get.return_value = _builder
    expected_job_status = JobStatus(
        id="job_id",
        cluster="cluster",
        name="foo",
        status="canceled",
        metadata="bar",
        done=False,
        state=JobState.PENDING,
    )
    mock_cloud.get_cluster.return_value = mock_cluster
    mock_cluster.cancel_job.return_value = expected_job_status
    launcher = Launcher()
    result = launcher.cancel("1", "cloud", "cluster")
    mock_cloud.get_cluster.assert_called_once_with("cluster")
    mock_cluster.cancel_job.assert_called_once_with("1")
    assert result == expected_job_status


def test_launcher_cancel_fails(mock_registry):
    with pytest.raises(ValueError) as exception_info:
        mock_cloud = Mock(spec=BaseCloud)

        def _builder():
            return mock_cloud

        mock_registry.get_all.return_value = {
            "cloud": _builder,
        }
        mock_registry.get.return_value = _builder
        mock_cloud.get_cluster.return_value = None
        launcher = Launcher()
        launcher.cancel("1", "cloud", "cluster")
    assert "not found" in str(exception_info.value)


def test_launcher_down_succeeds(mock_registry):
    mock_cluster = Mock(spec=BaseCluster)
    mock_cloud = Mock(spec=BaseCloud)

    def _builder():
        return mock_cloud

    mock_registry.get_all.return_value = {
        "cloud": _builder,
    }
    mock_registry.get.return_value = _builder
    mock_cloud.get_cluster.return_value = mock_cluster
    launcher = Launcher()
    launcher.down("cloud", "cluster")
    mock_cloud.get_cluster.assert_called_once_with("cluster")
    mock_cluster.down.assert_called_once()


def test_launcher_down_fails(mock_registry):
    with pytest.raises(ValueError) as exception_info:
        mock_cloud = Mock(spec=BaseCloud)

        def _builder():
            return mock_cloud

        mock_registry.get_all.return_value = {
            "cloud": _builder,
        }
        mock_registry.get.return_value = _builder
        mock_cloud.get_cluster.return_value = None
        launcher = Launcher()
        launcher.down("cloud", "cluster")
    assert "not found" in str(exception_info.value)


def test_launcher_status_multiple_clouds(mock_registry):
    sky_mock = Mock(spec=BaseCloud)
    polaris_mock = Mock(spec=BaseCloud)
    custom_mock = Mock(spec=BaseCloud)

    def _sky_builder():
        return sky_mock

    def _polaris_builder():
        return polaris_mock

    def _custom_builder():
        return custom_mock

    mock_registry.get_all.return_value = {
        "sky": _sky_builder,
        "polaris": _polaris_builder,
        "custom": _custom_builder,
    }
    mock_sky_cluster1 = Mock(spec=BaseCluster)
    mock_sky_cluster1.get_jobs.return_value = [
        JobStatus(
            id="1",
            cluster="sky1",
            name="foo",
            status="running",
            metadata="bar",
            done=False,
            state=JobState.PENDING,
        ),
        JobStatus(
            id="2",
            cluster="sky1",
            name="foo",
            status="running",
            metadata="bar",
            done=False,
            state=JobState.PENDING,
        ),
    ]
    mock_sky_cluster2 = Mock(spec=BaseCluster)
    mock_sky_cluster2.get_jobs.return_value = [
        JobStatus(
            id="1",
            cluster="sky2",
            name="foo",
            status="running",
            metadata="bar",
            done=False,
            state=JobState.PENDING,
        ),
    ]
    mock_polaris_cluster1 = Mock(spec=BaseCluster)
    mock_polaris_cluster1.get_jobs.return_value = [
        JobStatus(
            id="1",
            cluster="polaris1",
            name="foo",
            status="running",
            metadata="bar",
            done=False,
            state=JobState.PENDING,
        ),
    ]
    mock_polaris_cluster2 = Mock(spec=BaseCluster)
    mock_polaris_cluster2.get_jobs.return_value = []
    mock_polaris_cluster3 = Mock(spec=BaseCluster)
    mock_polaris_cluster3.get_jobs.return_value = [
        JobStatus(
            id="1",
            cluster="polaris3",
            name="foo",
            status="running",
            metadata="bar",
            done=False,
            state=JobState.PENDING,
        ),
    ]
    sky_mock.list_clusters.return_value = [mock_sky_cluster1, mock_sky_cluster2]
    polaris_mock.list_clusters.return_value = [
        mock_polaris_cluster1,
        mock_polaris_cluster2,
        mock_polaris_cluster3,
    ]
    custom_mock.list_clusters.return_value = []
    launcher = Launcher()
    statuses = launcher.status()
    assert statuses == {
        "custom": [],
        "sky": [
            JobStatus(
                id="1",
                cluster="sky1",
                name="foo",
                status="running",
                metadata="bar",
                done=False,
                state=JobState.PENDING,
            ),
            JobStatus(
                id="2",
                cluster="sky1",
                name="foo",
                status="running",
                metadata="bar",
                done=False,
                state=JobState.PENDING,
            ),
            JobStatus(
                id="1",
                cluster="sky2",
                name="foo",
                status="running",
                metadata="bar",
                done=False,
                state=JobState.PENDING,
            ),
        ],
        "polaris": [
            JobStatus(
                id="1",
                cluster="polaris1",
                name="foo",
                status="running",
                metadata="bar",
                done=False,
                state=JobState.PENDING,
            ),
            JobStatus(
                id="1",
                cluster="polaris3",
                name="foo",
                status="running",
                metadata="bar",
                done=False,
                state=JobState.PENDING,
            ),
        ],
    }


def test_launcher_status_filters_clusters(mock_registry):
    sky_mock = Mock(spec=BaseCloud)
    polaris_mock = Mock(spec=BaseCloud)
    custom_mock = Mock(spec=BaseCloud)

    def _sky_builder():
        return sky_mock

    def _polaris_builder():
        return polaris_mock

    def _custom_builder():
        return custom_mock

    mock_registry.get_all.return_value = {
        "sky": _sky_builder,
        "polaris": _polaris_builder,
        "custom": _custom_builder,
    }
    mock_sky_cluster1 = Mock(spec=BaseCluster)
    mock_sky_cluster1.name.return_value = "sky1"
    mock_sky_cluster1.get_jobs.return_value = [
        JobStatus(
            id="1",
            cluster="sky1",
            name="foo",
            status="running",
            metadata="bar",
            done=False,
            state=JobState.PENDING,
        ),
        JobStatus(
            id="2",
            cluster="sky1",
            name="foo",
            status="running",
            metadata="bar",
            done=False,
            state=JobState.PENDING,
        ),
    ]
    mock_sky_cluster2 = Mock(spec=BaseCluster)
    mock_sky_cluster2.name.return_value = "sky2"
    mock_sky_cluster2.get_jobs.return_value = [
        JobStatus(
            id="1",
            cluster="sky2",
            name="foo",
            status="running",
            metadata="bar",
            done=False,
            state=JobState.PENDING,
        ),
    ]
    mock_polaris_cluster1 = Mock(spec=BaseCluster)
    mock_polaris_cluster1.name.return_value = "polaris1"
    mock_polaris_cluster1.get_jobs.return_value = [
        JobStatus(
            id="1",
            cluster="polaris1",
            name="foo",
            status="running",
            metadata="bar",
            done=False,
            state=JobState.PENDING,
        ),
    ]
    mock_polaris_cluster2 = Mock(spec=BaseCluster)
    mock_polaris_cluster2.name.return_value = "polaris2"
    mock_polaris_cluster2.get_jobs.return_value = []
    mock_polaris_cluster3 = Mock(spec=BaseCluster)
    mock_polaris_cluster3.name.return_value = "polaris3"
    mock_polaris_cluster3.get_jobs.return_value = [
        JobStatus(
            id="1",
            cluster="polaris3",
            name="foo",
            status="running",
            metadata="bar",
            done=False,
            state=JobState.PENDING,
        ),
    ]
    sky_mock.list_clusters.return_value = [mock_sky_cluster1, mock_sky_cluster2]
    polaris_mock.list_clusters.return_value = [
        mock_polaris_cluster1,
        mock_polaris_cluster2,
        mock_polaris_cluster3,
    ]
    custom_mock.list_clusters.return_value = []
    launcher = Launcher()
    statuses = launcher.status(cluster="sky1")
    assert statuses == {
        "custom": [],
        "sky": [
            JobStatus(
                id="1",
                cluster="sky1",
                name="foo",
                status="running",
                metadata="bar",
                done=False,
                state=JobState.PENDING,
            ),
            JobStatus(
                id="2",
                cluster="sky1",
                name="foo",
                status="running",
                metadata="bar",
                done=False,
                state=JobState.PENDING,
            ),
        ],
        "polaris": [],
    }


def test_launcher_status_filters_jobs(mock_registry):
    sky_mock = Mock(spec=BaseCloud)
    polaris_mock = Mock(spec=BaseCloud)
    custom_mock = Mock(spec=BaseCloud)

    def _sky_builder():
        return sky_mock

    def _polaris_builder():
        return polaris_mock

    def _custom_builder():
        return custom_mock

    mock_registry.get_all.return_value = {
        "sky": _sky_builder,
        "polaris": _polaris_builder,
        "custom": _custom_builder,
    }
    mock_sky_cluster1 = Mock(spec=BaseCluster)
    mock_sky_cluster1.get_jobs.return_value = [
        JobStatus(
            id="1",
            cluster="sky1",
            name="foo",
            status="running",
            metadata="bar",
            done=False,
            state=JobState.PENDING,
        ),
        JobStatus(
            id="2",
            cluster="sky1",
            name="foo",
            status="running",
            metadata="bar",
            done=False,
            state=JobState.PENDING,
        ),
    ]
    mock_sky_cluster2 = Mock(spec=BaseCluster)
    mock_sky_cluster2.get_jobs.return_value = [
        JobStatus(
            id="1",
            cluster="sky2",
            name="foo",
            status="running",
            metadata="bar",
            done=False,
            state=JobState.PENDING,
        ),
    ]
    mock_polaris_cluster1 = Mock(spec=BaseCluster)
    mock_polaris_cluster1.get_jobs.return_value = [
        JobStatus(
            id="1",
            cluster="polaris1",
            name="foo",
            status="running",
            metadata="bar",
            done=False,
            state=JobState.PENDING,
        ),
    ]
    mock_polaris_cluster2 = Mock(spec=BaseCluster)
    mock_polaris_cluster2.get_jobs.return_value = []
    mock_polaris_cluster3 = Mock(spec=BaseCluster)
    mock_polaris_cluster3.get_jobs.return_value = [
        JobStatus(
            id="1",
            cluster="polaris3",
            name="foo",
            status="running",
            metadata="bar",
            done=False,
            state=JobState.PENDING,
        ),
    ]
    sky_mock.list_clusters.return_value = [mock_sky_cluster1, mock_sky_cluster2]
    polaris_mock.list_clusters.return_value = [
        mock_polaris_cluster1,
        mock_polaris_cluster2,
        mock_polaris_cluster3,
    ]
    custom_mock.list_clusters.return_value = []
    launcher = Launcher()
    statuses = launcher.status(id="1")
    assert statuses == {
        "custom": [],
        "sky": [
            JobStatus(
                id="1",
                cluster="sky1",
                name="foo",
                status="running",
                metadata="bar",
                done=False,
                state=JobState.PENDING,
            ),
            JobStatus(
                id="1",
                cluster="sky2",
                name="foo",
                status="running",
                metadata="bar",
                done=False,
                state=JobState.PENDING,
            ),
        ],
        "polaris": [
            JobStatus(
                id="1",
                cluster="polaris1",
                name="foo",
                status="running",
                metadata="bar",
                done=False,
                state=JobState.PENDING,
            ),
            JobStatus(
                id="1",
                cluster="polaris3",
                name="foo",
                status="running",
                metadata="bar",
                done=False,
                state=JobState.PENDING,
            ),
        ],
    }


def test_launcher_status_filters_clouds(mock_registry):
    sky_mock = Mock(spec=BaseCloud)
    polaris_mock = Mock(spec=BaseCloud)
    custom_mock = Mock(spec=BaseCloud)

    def _sky_builder():
        return sky_mock

    def _polaris_builder():
        return polaris_mock

    def _custom_builder():
        return custom_mock

    mock_registry.get_all.return_value = {
        "sky": _sky_builder,
        "polaris": _polaris_builder,
        "custom": _custom_builder,
    }
    mock_sky_cluster1 = Mock(spec=BaseCluster)
    mock_sky_cluster1.get_jobs.return_value = [
        JobStatus(
            id="1",
            cluster="sky1",
            name="foo",
            status="running",
            metadata="bar",
            done=False,
            state=JobState.PENDING,
        ),
        JobStatus(
            id="2",
            cluster="sky1",
            name="foo",
            status="running",
            metadata="bar",
            done=False,
            state=JobState.PENDING,
        ),
    ]
    mock_sky_cluster2 = Mock(spec=BaseCluster)
    mock_sky_cluster2.get_jobs.return_value = [
        JobStatus(
            id="1",
            cluster="sky2",
            name="foo",
            status="running",
            metadata="bar",
            done=False,
            state=JobState.PENDING,
        ),
    ]
    mock_polaris_cluster1 = Mock(spec=BaseCluster)
    mock_polaris_cluster1.get_jobs.return_value = [
        JobStatus(
            id="1",
            cluster="polaris1",
            name="foo",
            status="running",
            metadata="bar",
            done=False,
            state=JobState.PENDING,
        ),
    ]
    mock_polaris_cluster2 = Mock(spec=BaseCluster)
    mock_polaris_cluster2.get_jobs.return_value = []
    mock_polaris_cluster3 = Mock(spec=BaseCluster)
    mock_polaris_cluster3.get_jobs.return_value = [
        JobStatus(
            id="1",
            cluster="polaris3",
            name="foo",
            status="running",
            metadata="bar",
            done=False,
            state=JobState.PENDING,
        ),
    ]
    sky_mock.list_clusters.return_value = [mock_sky_cluster1, mock_sky_cluster2]
    polaris_mock.list_clusters.return_value = [
        mock_polaris_cluster1,
        mock_polaris_cluster2,
        mock_polaris_cluster3,
    ]
    custom_mock.list_clusters.return_value = []
    launcher = Launcher()
    statuses = launcher.status(cloud="sky")
    assert statuses == {
        "sky": [
            JobStatus(
                id="1",
                cluster="sky1",
                name="foo",
                status="running",
                metadata="bar",
                done=False,
                state=JobState.PENDING,
            ),
            JobStatus(
                id="2",
                cluster="sky1",
                name="foo",
                status="running",
                metadata="bar",
                done=False,
                state=JobState.PENDING,
            ),
            JobStatus(
                id="1",
                cluster="sky2",
                name="foo",
                status="running",
                metadata="bar",
                done=False,
                state=JobState.PENDING,
            ),
        ],
    }


def test_launcher_status_all_filters(mock_registry):
    sky_mock = Mock(spec=BaseCloud)
    polaris_mock = Mock(spec=BaseCloud)
    custom_mock = Mock(spec=BaseCloud)

    def _sky_builder():
        return sky_mock

    def _polaris_builder():
        return polaris_mock

    def _custom_builder():
        return custom_mock

    mock_registry.get_all.return_value = {
        "sky": _sky_builder,
        "polaris": _polaris_builder,
        "custom": _custom_builder,
    }
    mock_sky_cluster1 = Mock(spec=BaseCluster)
    mock_sky_cluster1.name.return_value = "foo"
    mock_sky_cluster1.get_jobs.return_value = [
        JobStatus(
            id="1",
            cluster="foo",
            name="foo",
            status="running",
            metadata="bar",
            done=False,
            state=JobState.PENDING,
        ),
        JobStatus(
            id="2",
            cluster="foo",
            name="foo",
            status="running",
            metadata="bar",
            done=False,
            state=JobState.PENDING,
        ),
    ]
    mock_sky_cluster2 = Mock(spec=BaseCluster)
    mock_sky_cluster2.name.return_value = "sky2"
    mock_sky_cluster2.get_jobs.return_value = [
        JobStatus(
            id="1",
            cluster="sky2",
            name="foo",
            status="running",
            metadata="bar",
            done=False,
            state=JobState.PENDING,
        ),
    ]
    mock_polaris_cluster1 = Mock(spec=BaseCluster)
    mock_polaris_cluster1.name.return_value = "foo"
    mock_polaris_cluster1.get_jobs.return_value = [
        JobStatus(
            id="1",
            cluster="foo",
            name="foo",
            status="running",
            metadata="bar",
            done=False,
            state=JobState.PENDING,
        ),
    ]
    mock_polaris_cluster2 = Mock(spec=BaseCluster)
    mock_polaris_cluster2.name.return_value = "polaris2"
    mock_polaris_cluster2.get_jobs.return_value = []
    mock_polaris_cluster3 = Mock(spec=BaseCluster)
    mock_polaris_cluster3.name.return_value = "polaris3"
    mock_polaris_cluster3.get_jobs.return_value = [
        JobStatus(
            id="1",
            cluster="polaris3",
            name="foo",
            status="running",
            metadata="bar",
            done=False,
            state=JobState.PENDING,
        ),
    ]
    sky_mock.list_clusters.return_value = [mock_sky_cluster1, mock_sky_cluster2]
    polaris_mock.list_clusters.return_value = [
        mock_polaris_cluster1,
        mock_polaris_cluster2,
        mock_polaris_cluster3,
    ]
    custom_mock.list_clusters.return_value = []
    launcher = Launcher()
    statuses = launcher.status(cluster="foo", cloud="sky", id="1")
    assert statuses == {
        "sky": [
            JobStatus(
                id="1",
                cluster="foo",
                name="foo",
                status="running",
                metadata="bar",
                done=False,
                state=JobState.PENDING,
            ),
        ],
    }


def test_launcher_status_inits_new_clouds(mock_registry):
    sky_mock = Mock(spec=BaseCloud)
    polaris_mock = Mock(spec=BaseCloud)
    custom_mock = Mock(spec=BaseCloud)

    def _sky_builder():
        return sky_mock

    def _polaris_builder():
        return polaris_mock

    def _custom_builder():
        return custom_mock

    mock_registry.get_all.side_effect = [
        {},
        {
            "sky": _sky_builder,
            "polaris": _polaris_builder,
            "custom": _custom_builder,
        },
    ]
    mock_sky_cluster1 = Mock(spec=BaseCluster)
    mock_sky_cluster1.get_jobs.return_value = [
        JobStatus(
            id="1",
            cluster="sky1",
            name="foo",
            status="running",
            metadata="bar",
            done=False,
            state=JobState.PENDING,
        ),
        JobStatus(
            id="2",
            cluster="sky1",
            name="foo",
            status="running",
            metadata="bar",
            done=False,
            state=JobState.PENDING,
        ),
    ]
    mock_sky_cluster2 = Mock(spec=BaseCluster)
    mock_sky_cluster2.get_jobs.return_value = [
        JobStatus(
            id="1",
            cluster="sky2",
            name="foo",
            status="running",
            metadata="bar",
            done=False,
            state=JobState.PENDING,
        ),
    ]
    mock_polaris_cluster1 = Mock(spec=BaseCluster)
    mock_polaris_cluster1.get_jobs.return_value = [
        JobStatus(
            id="1",
            cluster="polaris1",
            name="foo",
            status="running",
            metadata="bar",
            done=False,
            state=JobState.PENDING,
        ),
    ]
    mock_polaris_cluster2 = Mock(spec=BaseCluster)
    mock_polaris_cluster2.get_jobs.return_value = []
    mock_polaris_cluster3 = Mock(spec=BaseCluster)
    mock_polaris_cluster3.get_jobs.return_value = [
        JobStatus(
            id="1",
            cluster="polaris3",
            name="foo",
            status="running",
            metadata="bar",
            done=False,
            state=JobState.PENDING,
        ),
    ]
    sky_mock.list_clusters.return_value = [mock_sky_cluster1, mock_sky_cluster2]
    polaris_mock.list_clusters.return_value = [
        mock_polaris_cluster1,
        mock_polaris_cluster2,
        mock_polaris_cluster3,
    ]
    custom_mock.list_clusters.return_value = []
    launcher = Launcher()
    statuses = launcher.status()
    # On the first call, statuses should be empty.
    assert statuses == {}
    # On the second call, we've registered new clouds that yield jobs.
    new_statuses = launcher.status()
    assert new_statuses == {
        "custom": [],
        "sky": [
            JobStatus(
                id="1",
                cluster="sky1",
                name="foo",
                status="running",
                metadata="bar",
                done=False,
                state=JobState.PENDING,
            ),
            JobStatus(
                id="2",
                cluster="sky1",
                name="foo",
                status="running",
                metadata="bar",
                done=False,
                state=JobState.PENDING,
            ),
            JobStatus(
                id="1",
                cluster="sky2",
                name="foo",
                status="running",
                metadata="bar",
                done=False,
                state=JobState.PENDING,
            ),
        ],
        "polaris": [
            JobStatus(
                id="1",
                cluster="polaris1",
                name="foo",
                status="running",
                metadata="bar",
                done=False,
                state=JobState.PENDING,
            ),
            JobStatus(
                id="1",
                cluster="polaris3",
                name="foo",
                status="running",
                metadata="bar",
                done=False,
                state=JobState.PENDING,
            ),
        ],
    }


def test_launcher_stop_succeeds(mock_registry):
    mock_cluster = Mock(spec=BaseCluster)
    mock_cloud = Mock(spec=BaseCloud)

    def _builder():
        return mock_cloud

    mock_registry.get_all.return_value = {
        "cloud": _builder,
    }
    mock_registry.get.return_value = _builder
    mock_cloud.get_cluster.return_value = mock_cluster
    launcher = Launcher()
    launcher.stop("cloud", "cluster")
    mock_cloud.get_cluster.assert_called_once_with("cluster")
    mock_cluster.stop.assert_called_once()


def test_launcher_stop_fails(mock_registry):
    with pytest.raises(ValueError) as exception_info:
        mock_cloud = Mock(spec=BaseCloud)

        def _builder():
            return mock_cloud

        mock_registry.get_all.return_value = {
            "cloud": _builder,
        }
        mock_registry.get.return_value = _builder
        mock_cloud.get_cluster.return_value = None
        launcher = Launcher()
        launcher.stop("cloud", "cluster")
    assert "not found" in str(exception_info.value)


def test_launcher_which_clouds_updates_over_time(mock_registry):
    sky_mock = Mock(spec=BaseCloud)
    polaris_mock = Mock(spec=BaseCloud)
    custom_mock = Mock(spec=BaseCloud)

    def _sky_builder():
        return sky_mock

    def _polaris_builder():
        return polaris_mock

    def _custom_builder():
        return custom_mock

    mock_registry.get_all.side_effect = [
        {
            "sky": _sky_builder,
        },
        {
            "polaris": _polaris_builder,
        },
        {
            "sky": _sky_builder,
            "polaris": _polaris_builder,
            "custom": _custom_builder,
        },
    ]
    launcher = Launcher()
    assert launcher.which_clouds() == ["sky"]
    assert launcher.which_clouds() == ["polaris"]
    assert launcher.which_clouds() == ["sky", "polaris", "custom"]


def test_launcher_does_not_initialize_clouds_eagerly(mock_registry):
    """Verify clouds are not initialized in __init__."""
    launcher = Launcher()

    # Clouds dict should be empty after construction
    assert len(launcher._clouds) == 0


def test_launcher_initializes_cloud_on_first_use(mock_registry):
    """Verify clouds are initialized lazily on first access."""
    mock_cloud = Mock(spec=BaseCloud)
    cloud_builder = Mock(return_value=mock_cloud)

    mock_registry.get.return_value = cloud_builder

    launcher = Launcher()
    assert len(launcher._clouds) == 0  # Not initialized yet

    # Access the cloud for the first time
    result = launcher.get_cloud("sky")

    # Now it should be initialized
    from oumi.core.registry import RegistryType

    mock_registry.get.assert_called_once_with("sky", RegistryType.CLOUD)
    cloud_builder.assert_called_once()  # Builder was called
    assert result == mock_cloud
    assert len(launcher._clouds) == 1
    assert launcher._clouds["sky"] == mock_cloud


def test_launcher_reuses_initialized_cloud(mock_registry):
    """Verify clouds are not re-initialized on subsequent accesses."""
    mock_cloud = Mock(spec=BaseCloud)
    cloud_builder = Mock(return_value=mock_cloud)

    mock_registry.get.return_value = cloud_builder

    launcher = Launcher()

    # First access - should initialize
    result1 = launcher.get_cloud("sky")
    assert mock_registry.get.call_count == 1
    assert cloud_builder.call_count == 1

    # Second access - should reuse
    result2 = launcher.get_cloud("sky")
    assert mock_registry.get.call_count == 1  # Still 1, not called again!
    assert cloud_builder.call_count == 1  # Still 1, not called again!
    assert result1 is result2  # Same instance


def test_launcher_initializes_multiple_clouds_independently(mock_registry):
    """Verify multiple clouds can be initialized independently."""
    sky_cloud = Mock(spec=BaseCloud)
    polaris_cloud = Mock(spec=BaseCloud)

    def mock_get(name, reg_type):
        if name == "sky":
            return lambda: sky_cloud
        elif name == "polaris":
            return lambda: polaris_cloud
        return None

    mock_registry.get.side_effect = mock_get

    launcher = Launcher()
    assert len(launcher._clouds) == 0

    # Initialize sky cloud
    result1 = launcher.get_cloud("sky")
    assert len(launcher._clouds) == 1
    assert result1 == sky_cloud

    # Initialize polaris cloud
    result2 = launcher.get_cloud("polaris")
    assert len(launcher._clouds) == 2
    assert result2 == polaris_cloud

    # Both should be cached
    assert launcher._clouds["sky"] == sky_cloud
    assert launcher._clouds["polaris"] == polaris_cloud


def test_launcher_initialize_new_clouds_adds_only_new_clouds(mock_registry):
    """Verify _initialize_new_clouds only adds clouds not already cached."""
    existing_cloud = Mock(spec=BaseCloud)
    new_cloud = Mock(spec=BaseCloud)

    launcher = Launcher()
    launcher._clouds["existing"] = existing_cloud  # Manually add

    mock_registry.get_all.return_value = {
        "existing": lambda: Mock(spec=BaseCloud),  # Different instance
        "new": lambda: new_cloud,
    }

    launcher._initialize_new_clouds()

    # Existing cloud should NOT be replaced
    assert launcher._clouds["existing"] is existing_cloud
    # New cloud should be added
    assert launcher._clouds["new"] is new_cloud
    assert len(launcher._clouds) == 2


def test_launcher_export_methods(mock_registry):
    assert LAUNCHER.up == up
    assert LAUNCHER.run == run
    assert LAUNCHER.cancel == cancel
    assert LAUNCHER.down == down
    assert LAUNCHER.status == status
    assert LAUNCHER.stop == stop
    assert LAUNCHER.get_cloud == get_cloud
    assert LAUNCHER.which_clouds == which_clouds


def _verify_no_extra_import(extra_module: str):
    """Verifies that extra modules are not imported."""
    import sys

    import oumi.launcher  # noqa

    assert extra_module not in sys.modules, f"{extra_module} was imported."


def test_launcher_no_sky_dependency():
    # Ensure that sky is lazy loaded so it doesn't cause DB contention in multinode
    # jobs: https://github.com/oumi-ai/oumi/issues/1605

    set_start_method("spawn", force=True)
    process = Process(target=_verify_no_extra_import, args=["sky"])
    process.start()
    process.join()
    assert process.exitcode == 0, (
        "Sky was imported as part of the launcher module. This is a regression."
    )



# Copyright 2025 - Jiayi Pan
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Derived from https://github.com/Jiayi-Pan/TinyZero/blob/main/verl/utils/reward_score/countdown.py.

This file was slightly modified to be an Oumi reward registry function.
"""

import re
from typing import Any, Optional

from oumi.core.registry import RegistryType, register


def _extract_solution(solution_str: str) -> Optional[str]:
    """Extracts the equation from the solution string.

    Args:
        solution_str: The response from the LLM.

    Returns:
        The equation from the solution string, or None if not found.
    """
    solution_str = solution_str.split("\n")[-1]

    answer_pattern = r"<answer>(.*?)</answer>"
    match = re.finditer(answer_pattern, solution_str)
    matches = list(match)
    if matches:
        final_answer = matches[-1].group(1).strip()
    else:
        final_answer = None
    return final_answer


def _validate_equation(equation_str: str, available_numbers: list[int]) -> bool:
    """Validates that equation only uses available numbers and each number once.

    Args:
        equation_str: The equation to validate.
        available_numbers: The list of available numbers.

    Returns:
        True if the equation uses each available number exactly once, else False.
    """
    try:
        # Extract all numbers from the equation
        numbers_in_eq = [int(n) for n in re.findall(r"\d+", equation_str)]

        # Check if all numbers in equation are available
        available_numbers = sorted(available_numbers)
        numbers_in_eq = sorted(numbers_in_eq)

        # Each number should be used exactly once
        return numbers_in_eq == available_numbers
    except Exception:
        return False


def _evaluate_equation(equation_str: str) -> Optional[float]:
    """Safely evaluates the arithmetic equation using eval() with precautions."""
    try:
        # Regex that only allows numbers, operators, parentheses and whitespace
        allowed_pattern = r"^[\d+\-*/().\s]+$"
        if not re.match(allowed_pattern, equation_str):
            raise ValueError("Invalid characters in equation.")

        # Evaluate the equation with restricted globals and locals
        result = eval(equation_str, {"__builtins__": None}, {})
        return result
    except Exception:
        return None


@register("countdown", RegistryType.REWARD_FUNCTION)
def countdown_reward(
    data_source: str,
    solution_str: str,
    ground_truth: dict[str, Any],
    extra_info: dict[str, Any],
    format_score=0.0,
    score=1.0,
) -> float:
    """Custom reward function for the Countdown task.

    Currently, this function only works with the VERL_GRPO trainer.

    Args:
        data_source: The data source.
        solution_str: The response from the LLM.
        ground_truth: Dictionary containing target number and available numbers
        extra_info: Extra information about the sample.
        format_score: The score for correct format but wrong answer.
        score: The score for the correct answer.

    Returns:
        `score` if the equation is valid and correct,
        `format_score` if the answer was parsed properly but the equation is incorrect,
        `0` if the answer was not parsed properly.
    """
    target = ground_truth["target"]
    numbers = ground_truth["numbers"]

    equation = _extract_solution(solution_str=solution_str)

    if equation is None:
        return 0

    # Validate equation uses correct numbers
    if not _validate_equation(equation, numbers):
        return format_score

    # Evaluate equation
    try:
        result = _evaluate_equation(equation)
        if result is None:
            return format_score

        if abs(result - target) < 1e-5:  # Account for floating point precision
            return score
        else:
            return format_score
    except Exception:
        return format_score



import pytest
import typer
from typer.testing import CliRunner

from oumi.cli.env import env

runner = CliRunner()


#
# Fixtures
#
@pytest.fixture
def app():
    fake_app = typer.Typer()
    fake_app.command()(env)
    yield fake_app


def test_env_runs_without_exceptions(app):
    result = runner.invoke(app, [])
    assert result.exit_code == 0
    assert "Oumi environment information:" in result.stdout



from unittest.mock import patch

import pytest

from oumi.core.configs import GenerationParams, ModelParams, RemoteParams
from oumi.core.types.conversation import Conversation, Message, Role
from oumi.inference.anthropic_inference_engine import AnthropicInferenceEngine


@pytest.fixture
def anthropic_engine():
    return AnthropicInferenceEngine(
        model_params=ModelParams(model_name="claude-3"),
        remote_params=RemoteParams(api_key="test_api_key", api_url="<placeholder>"),
    )


def test_convert_conversation_to_api_input(anthropic_engine):
    conversation = Conversation(
        messages=[
            Message(content="System message", role=Role.SYSTEM),
            Message(content="User message", role=Role.USER),
            Message(content="Assistant message", role=Role.ASSISTANT),
        ]
    )
    generation_params = GenerationParams(max_new_tokens=100)

    result = anthropic_engine._convert_conversation_to_api_input(
        conversation, generation_params, anthropic_engine._model_params
    )

    assert result["model"] == "claude-3"
    assert result["system"] == "System message"
    assert len(result["messages"]) == 2
    assert result["messages"][0]["content"] == "User message"
    assert result["messages"][0]["role"] == "user"
    assert result["messages"][1]["content"] == "Assistant message"
    assert result["messages"][1]["role"] == "assistant"
    assert result["max_tokens"] == 100


def test_convert_api_output_to_conversation(anthropic_engine):
    original_conversation = Conversation(
        messages=[
            Message(content="User message", role=Role.USER),
        ],
        metadata={"key": "value"},
        conversation_id="test_id",
    )
    api_response = {"content": [{"text": "Assistant response"}]}

    result = anthropic_engine._convert_api_output_to_conversation(
        api_response, original_conversation
    )

    assert len(result.messages) == 2
    assert result.messages[0].content == "User message"
    assert result.messages[1].content == "Assistant response"
    assert result.messages[1].role == Role.ASSISTANT
    assert result.metadata == {"key": "value"}
    assert result.conversation_id == "test_id"


def test_get_request_headers(anthropic_engine):
    remote_params = RemoteParams(api_key="test_api_key", api_url="<placeholder>")

    with patch.object(
        AnthropicInferenceEngine,
        "_get_api_key",
        return_value="test_api_key",
    ):
        result = anthropic_engine._get_request_headers(remote_params)

    assert result["Content-Type"] == "application/json"
    assert result["anthropic-version"] == AnthropicInferenceEngine.anthropic_version
    assert result["X-API-Key"] == "test_api_key"


def test_remote_params_defaults():
    anthropic_engine = AnthropicInferenceEngine(
        model_params=ModelParams(model_name="some_model"),
    )
    assert anthropic_engine._remote_params.num_workers == 5
    assert anthropic_engine._remote_params.politeness_policy == 60.0



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
from importlib.util import find_spec
from pathlib import Path
from typing import Final, Optional, Union

import PIL.Image
import requests

from oumi.utils.logging import logger

# For details on image modes, see
# https://pillow.readthedocs.io/en/stable/handbook/concepts.html#modes
DEFAULT_IMAGE_MODE: Final[str] = "RGB"

_FILE_URL_PREFIX: Final[str] = "file://"
_DEFAULT_PDF_DPI: Final[int] = 200


def create_png_bytes_from_image(pil_image: PIL.Image.Image) -> bytes:
    """Encodes PIL image into PNG format, and returns PNG image bytes.

    Args:
        pil_image: An input image.

    Returns:
        bytes: PNG bytes representation of the image.
    """
    try:
        output = io.BytesIO()
        pil_image.save(output, format="PNG")
        return output.getvalue()
    except Exception:
        logger.error("Failed to convert an image to PNG bytes.")
        raise


def create_png_bytes_from_image_list(pil_images: list[PIL.Image.Image]) -> list[bytes]:
    """Encodes PIL images into PNG format, and returns PNG image bytes.

    Args:
        pil_images: A list of input images.

    Returns:
        A list of PNG-encoded images.
    """
    return [create_png_bytes_from_image(image) for image in pil_images]


def convert_pil_image_mode(
    image: PIL.Image.Image, *, mode: Optional[str]
) -> PIL.Image.Image:
    """Converts a PIL image to the requested mode (if it's not in that mode already) .

    Args:
        image: An input image.
        mode: The requested image mode e.g., "RGB", "HSV", "RGBA",
            "P" (8-bit pixels, using a color palette).
            For details, see https://pillow.readthedocs.io/en/stable/handbook/concepts.html#modes

    Returns:
        An image in the requested mode . If an input image was already in the correct
        mode then return it for efficiency.
        Otherwise, a different image object is returned.
    """
    if not mode or image.mode == mode:
        # Return the original object for better performance.
        return image

    old_mode = image.mode
    try:
        return image.convert(mode)
    except Exception as e:
        raise RuntimeError(
            f"Failed to convert an image from {old_mode} to {mode} mode!"
        ) from e


def load_pil_image_from_path(
    input_image_filepath: Union[str, Path], mode: str = DEFAULT_IMAGE_MODE
) -> PIL.Image.Image:
    """Loads an image from a path.

    Args:
        input_image_filepath: A file path of an image.
            The image can be in any format supported by PIL.
        mode: The requested image mode e.g., "RGB", "HSV", "RGBA",
            "P" (8-bit pixels, using a color palette).
            For details, see https://pillow.readthedocs.io/en/stable/handbook/concepts.html#modes

    Returns:
        bytes: PNG bytes representation of the image.
    """
    if not input_image_filepath:
        raise ValueError("Empty image file path.")

    if isinstance(
        input_image_filepath, str
    ) and input_image_filepath.lower().startswith(_FILE_URL_PREFIX):
        input_image_filepath = input_image_filepath[len(_FILE_URL_PREFIX) :]

    input_image_filepath = Path(input_image_filepath)
    if not input_image_filepath.is_file():
        raise ValueError(
            f"Image path is not a file: {input_image_filepath}"
            if input_image_filepath.exists()
            else f"Image path doesn't exist: {input_image_filepath}"
        )

    try:
        pil_image = convert_pil_image_mode(
            PIL.Image.open(input_image_filepath), mode=mode
        )
    except Exception:
        logger.error(f"Failed to load an image from path: {input_image_filepath}")
        raise
    return pil_image


def load_pil_image_from_url(
    input_image_url: str, mode: str = DEFAULT_IMAGE_MODE
) -> PIL.Image.Image:
    """Loads a PIL image from a URL.

    Args:
        input_image_url: An image URL.
        mode: The requested image mode e.g., "RGB", "HSV", "RGBA",
            "P" (8-bit pixels, using a color palette).
            For details, see https://pillow.readthedocs.io/en/stable/handbook/concepts.html#modes

    Returns:
        bytes: PNG bytes representation of the image.
    """
    if not input_image_url:
        raise ValueError("Empty image URL!")

    try:
        response = requests.get(input_image_url, stream=True)
        response.raise_for_status()
    except requests.exceptions.RequestException:
        logger.exception(f"Failed to download image: '{input_image_url}'")
        raise
    return load_pil_image_from_bytes(response.content, mode=mode)


def load_pil_image_from_bytes(
    image_bytes: Optional[bytes], mode: str = DEFAULT_IMAGE_MODE
) -> PIL.Image.Image:
    """Loads an image from raw image bytes.

    Args:
        image_bytes: A input image bytes. Can be in any image format supported by PIL.
        mode: The requested image mode e.g., "RGB", "HSV", "RGBA",
            "P" (8-bit pixels, using a color palette).
            For details, see https://pillow.readthedocs.io/en/stable/handbook/concepts.html#modes

    Returns:
        PIL.Image.Image: PIL representation of the image.
    """
    if image_bytes is None or len(image_bytes) == 0:
        raise ValueError("No image bytes.")

    try:
        pil_image = convert_pil_image_mode(
            PIL.Image.open(io.BytesIO(image_bytes)), mode=mode
        )
    except Exception:
        logger.error(
            f"Failed to load an image from raw image bytes ({len(image_bytes)} bytes)."
        )
        raise
    return pil_image


def _check_pdf2image_dependency():
    if not find_spec("pdf2image"):
        raise RuntimeError(
            "Failed to find the required dependency package: 'pdf2image'. "
            "Run `pip install oumi[file_formats]`, and try again."
        )


def load_pdf_pages_from_path(
    input_pdf_filepath: Union[str, Path],
    *,
    dpi: int = _DEFAULT_PDF_DPI,
    mode: str = DEFAULT_IMAGE_MODE,
) -> list[PIL.Image.Image]:
    """Loads PDF pages as PIL images from a path.

    Args:
        input_pdf_filepath: A file path of an PDF document.
        dpi: Resolution to use for PDF page images (dots per inch).
        mode: The requested image mode e.g., "RGB", "HSV", "RGBA",
            "P" (8-bit pixels, using a color palette).
            For details, see https://pillow.readthedocs.io/en/stable/handbook/concepts.html#modes

    Returns:
        PDF pages as PIL images (PIL.Image.Image).
    """
    if not input_pdf_filepath:
        raise ValueError("Empty PDF file path.")

    if isinstance(input_pdf_filepath, str) and input_pdf_filepath.lower().startswith(
        _FILE_URL_PREFIX
    ):
        input_pdf_filepath = input_pdf_filepath[len(_FILE_URL_PREFIX) :]

    input_filepath = Path(input_pdf_filepath)
    if not input_filepath.is_file():
        raise ValueError(
            f"PDF path is not a file: {input_filepath}"
            if input_filepath.exists()
            else f"PDF path doesn't exist: {input_filepath}"
        )

    _check_pdf2image_dependency()
    import pdf2image  # pyright: ignore[reportMissingImports]

    page_images = pdf2image.convert_from_path(input_filepath, dpi=dpi)
    num_pages = len(page_images)
    for page_idx in range(num_pages):
        try:
            page_images[page_idx] = convert_pil_image_mode(
                page_images[page_idx], mode=mode
            )
        except Exception:
            logger.error(
                "Failed to convert image mode for PDF page "
                f"{page_idx + 1} of {num_pages}: {input_filepath}"
            )
            raise
    return page_images


def load_pdf_pages_from_bytes(
    pdf_bytes: Optional[bytes],
    *,
    dpi: int = _DEFAULT_PDF_DPI,
    mode: str = DEFAULT_IMAGE_MODE,
) -> list[PIL.Image.Image]:
    """Loads PDF pages as PIL images from raw PDF file bytes.

    Args:
        pdf_bytes: PDF file content.
        dpi: Resolution to use for PDF page images (dots per inch).
        mode: The requested image mode e.g., "RGB", "HSV", "RGBA",
            "P" (8-bit pixels, using a color palette).
            For details, see https://pillow.readthedocs.io/en/stable/handbook/concepts.html#modes

    Returns:
        PDF pages as PIL images (PIL.Image.Image).
    """
    if pdf_bytes is None or len(pdf_bytes) == 0:
        raise ValueError("No PDF bytes.")

    _check_pdf2image_dependency()
    import pdf2image  # pyright: ignore[reportMissingImports]

    page_images = pdf2image.convert_from_bytes(pdf_bytes, dpi=dpi)
    num_pages = len(page_images)
    for page_idx in range(num_pages):
        try:
            page_images[page_idx] = convert_pil_image_mode(
                page_images[page_idx], mode=mode
            )
        except Exception:
            logger.error(
                "Failed to convert image mode for PDF page "
                f"{page_idx + 1} or {num_pages}"
            )
            raise
    return page_images


def load_pdf_pages_from_url(
    pdf_url: str, *, dpi: int = _DEFAULT_PDF_DPI, mode: str = DEFAULT_IMAGE_MODE
) -> list[PIL.Image.Image]:
    """Loads PDF pages as PIL images from from PDF URL.

    Args:
        pdf_url: A PDF URL.
        dpi: Resolution to use for PDF page images (dots per inch).
        mode: The requested image mode e.g., "RGB", "HSV", "RGBA",
            "P" (8-bit pixels, using a color palette).
            For details, see https://pillow.readthedocs.io/en/stable/handbook/concepts.html#modes

    Returns:
        PDF pages as PIL images (PIL.Image.Image).
    """
    if not pdf_url:
        raise ValueError("Empty PDF URL!")

    try:
        response = requests.get(pdf_url, stream=True)
        response.raise_for_status()
    except requests.exceptions.RequestException:
        logger.exception(f"Failed to download PDF: '{pdf_url}'")
        raise
    return load_pdf_pages_from_bytes(response.content, dpi=dpi, mode=mode)


def create_png_bytes_from_image_bytes(
    image_bytes: Optional[bytes], mode: str = DEFAULT_IMAGE_MODE
) -> bytes:
    """Loads an image from raw image bytes, and converts to PNG image bytes.

    Args:
        image_bytes: A input image bytes. Can be in any image format supported by PIL.
        mode: The requested image mode e.g., "RGB", "HSV", "RGBA",
            "P" (8-bit pixels, using a color palette).
            For details, see https://pillow.readthedocs.io/en/stable/handbook/concepts.html#modes

    Returns:
        bytes: PNG bytes representation of the image.
    """
    pil_image = load_pil_image_from_bytes(image_bytes, mode=mode)
    return create_png_bytes_from_image(pil_image)


def load_image_png_bytes_from_path(
    input_image_filepath: Union[str, Path], mode: str = DEFAULT_IMAGE_MODE
) -> bytes:
    """Loads an image from a path, converts it to PNG, and returns image bytes.

    Args:
        input_image_filepath: A file path of an image.
            The image can be in any format supported by PIL.
        mode: The requested image mode e.g., "RGB", "HSV", "RGBA",
            "P" (8-bit pixels, using a color palette).
            For details, see https://pillow.readthedocs.io/en/stable/handbook/concepts.html#modes

    Returns:
        bytes: PNG bytes representation of the image.
    """
    pil_image = load_pil_image_from_path(input_image_filepath, mode=mode)
    return create_png_bytes_from_image(pil_image)


def load_image_png_bytes_from_url(
    input_image_url: str, mode: str = DEFAULT_IMAGE_MODE
) -> bytes:
    """Loads an image from a URL, converts it to PNG, and returns image bytes.

    Args:
        input_image_url: An image URL.
        mode: The requested image mode e.g., "RGB", "HSV", "RGBA",
            "P" (8-bit pixels, using a color palette).
            For details, see https://pillow.readthedocs.io/en/stable/handbook/concepts.html#modes

    Returns:
        bytes: PNG bytes representation of the image.
    """
    pil_image = load_pil_image_from_url(input_image_url, mode=mode)
    return create_png_bytes_from_image(pil_image)



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Annotated

import typer
from rich.table import Table

import oumi.cli.cli_utils as cli_utils
from oumi.cli.alias import AliasType, try_get_config_name_for_alias
from oumi.utils.logging import logger


def evaluate(
    ctx: typer.Context,
    config: Annotated[
        str,
        typer.Option(
            *cli_utils.CONFIG_FLAGS, help="Path to the configuration file for training."
        ),
    ],
    level: cli_utils.LOG_LEVEL_TYPE = None,
    verbose: cli_utils.VERBOSE_TYPE = False,
):
    """Evaluate a model.

    Args:
        ctx: The Typer context object.
        config: Path to the configuration file for evaluation.
        level: The logging level for the specified command.
        verbose: Enable verbose logging with additional debug information.
    """
    extra_args = cli_utils.parse_extra_cli_args(ctx)

    config = str(
        cli_utils.resolve_and_fetch_config(
            try_get_config_name_for_alias(config, AliasType.EVAL),
        )
    )
    with cli_utils.CONSOLE.status(
        "[green]Loading configuration...[/green]", spinner="dots"
    ):
        # Delayed imports
        from oumi import evaluate as oumi_evaluate
        from oumi.core.configs import EvaluationConfig
        # End imports

    # Load configuration
    parsed_config: EvaluationConfig = EvaluationConfig.from_yaml_and_arg_list(
        config, extra_args, logger=logger
    )
    parsed_config.finalize_and_validate()

    if verbose:
        # Print configuration for verification
        parsed_config.print_config(logger)

    # Run evaluation
    with cli_utils.CONSOLE.status(
        "[green]Running evaluation...[/green]", spinner="dots"
    ):
        results = oumi_evaluate(parsed_config)
    # Make a best-effort attempt at parsing metrics.
    for task_result in results:
        table = Table(
            title="Evaluation Results",
            title_style="bold magenta",
            show_lines=True,
        )
        table.add_column("Benchmark", style="cyan")
        table.add_column("Metric", style="yellow")
        table.add_column("Score", style="green")
        table.add_column("Std Error", style="dim")
        parsed_results = task_result.get("results", {})
        if not isinstance(parsed_results, dict):
            continue
        for task_name, metrics in parsed_results.items():
            # Get the benchmark display name from our benchmarks list

            if not isinstance(metrics, dict):
                # Skip if the metrics are not in a dict format
                table.add_row(
                    task_name,
                    "<unknown>",
                    "<unknown>",
                    "-",
                )
                continue
            benchmark_name: str = metrics.get("alias", task_name)
            # Process metrics
            for metric_name, value in metrics.items():
                metric_name: str = str(metric_name)
                if isinstance(value, (int, float)):
                    # Extract base metric name and type
                    base_name, *metric_type = metric_name.split(",")

                    # Skip if this is a stderr metric
                    # we'll handle it with the main metric
                    if base_name.endswith("_stderr"):
                        continue

                    # Get corresponding stderr if it exists
                    stderr_key = f"{base_name}_stderr,{metric_type[0] if metric_type else 'none'}"  # noqa E501
                    stderr_value = metrics.get(stderr_key)
                    stderr_display = (
                        f"±{stderr_value:.2%}" if stderr_value is not None else "-"
                    )

                    # Clean up metric name
                    clean_metric = base_name.replace("_", " ").title()

                    if isinstance(value, float):
                        if value > 1:
                            value_str = f"{value:.2f}"
                        else:
                            value_str = f"{value:.2%}"
                    else:
                        # Includes ints
                        value_str = str(value)
                    table.add_row(
                        benchmark_name,
                        clean_metric,
                        value_str,
                        stderr_display,
                    )
        cli_utils.CONSOLE.print(table)



# Test for a non-textual custom model (CNNClassifier).
import tempfile
from pathlib import Path
from typing import Any, NamedTuple, Optional, Union

import numpy as np
import pandas as pd
import torch
from torch.utils.data import DataLoader
from typing_extensions import override

from oumi.builders import build_dataset, build_model
from oumi.core.configs import ModelParams, TrainingParams
from oumi.core.datasets import BaseMapDataset
from oumi.core.registry import REGISTRY, RegistryType, register_dataset
from oumi.core.trainers.oumi_trainer import Trainer
from oumi.utils.logging import logger


@register_dataset("npz_file")
class NpzDataset(BaseMapDataset):
    """Loads dataset from Numpy .npz archive."""

    default_dataset = "custom"

    def __init__(
        self,
        *,
        dataset_name: Optional[str] = None,
        dataset_path: Optional[Union[str, Path]] = None,
        split: Optional[str] = None,
        npz_split_col: Optional[str] = None,
        npz_allow_pickle: bool = False,
        **kwargs,
    ) -> None:
        """Initializes a new instance of the NpzDataset class.

        Args:
            dataset_name: Dataset name.
            dataset_path: Path to .npz file.
            split: Dataset split.
            npz_split_col: Name of '.npz' array containing dataset split info.
                If unspecified, then the name "split" is assumed by default.
            npz_allow_pickle: Whether pickle is allowed when loading data
                from the npz archive.
            **kwargs: Additional arguments to pass to the parent class.
        Raises:
            ValueError: If dataset_path is not provided, or
                if .npz file contains data in unexpected format.
        """
        if not dataset_path:
            raise ValueError("`dataset_path` must be provided")
        super().__init__(
            dataset_name=dataset_name,
            dataset_path=(str(dataset_path) if dataset_path is not None else None),
            split=split,
            **kwargs,
        )
        self._npz_allow_pickle = npz_allow_pickle
        self._npz_split_col = npz_split_col

        dataset_path = Path(dataset_path)
        if not dataset_path.is_file():
            raise ValueError(f"Path is not a file! '{dataset_path}'")
        elif dataset_path.suffix.lower() != ".npz":
            raise ValueError(f"File extension is not '.npz'! '{dataset_path}'")

        self._data = self._load_data()

    @staticmethod
    def _to_list(x: np.ndarray) -> list:
        # `pd.DataFrame` expects Python lists for columns
        # (elements can still be `ndarray`)
        if len(x.shape) > 1:
            return [x[i, ...] for i in range(x.shape[0])]
        return x.tolist()

    @override
    def _load_data(self) -> pd.DataFrame:
        data_dict = {}
        if not self.dataset_path:
            raise ValueError("dataset_path is empty!")
        with np.load(self.dataset_path, allow_pickle=self._npz_allow_pickle) as npzfile:
            feature_names = list(sorted(npzfile.files))
            if len(feature_names) == 0:
                raise ValueError(
                    f"'.npz' archive contains no data! '{self.dataset_path}'"
                )
            num_examples = None
            for feature_name in feature_names:
                col_data = npzfile[feature_name]
                assert isinstance(col_data, np.ndarray)
                if num_examples is None:
                    num_examples = col_data.shape[0]
                elif num_examples != col_data.shape[0]:
                    raise ValueError(
                        "Inconsistent number of examples for features "
                        f"'{feature_name}' and '{feature_names[0]}': "
                        f"{col_data.shape[0]} vs {num_examples}!"
                    )
                data_dict[feature_name] = self._to_list(col_data)

        logger.info(
            f"Loaded {num_examples} examples "
            f"with the features: {sorted(data_dict.keys())}!"
        )

        dataframe: pd.DataFrame = pd.DataFrame(data_dict)

        split_feature_name = (self._npz_split_col or "split") if self.split else None
        if split_feature_name:
            if split_feature_name not in dataframe:
                raise ValueError(
                    f"'.npz' doesn't contain data split info: '{split_feature_name}'!"
                )
            dataframe = pd.DataFrame(
                dataframe[dataframe[split_feature_name] == self.split].drop(
                    split_feature_name, axis=1
                ),
                copy=True,
            )
        return dataframe

    @override
    def transform(self, sample: pd.Series) -> dict:
        """Preprocesses the inputs in the given sample."""
        return sample.to_dict()


def _convert_example_to_model_input(example: dict, device: torch.device) -> dict:
    return {
        key: (
            torch.from_numpy(value)
            if isinstance(value, np.ndarray)
            else torch.from_numpy(np.asarray(value))
        ).to(device, non_blocking=True)
        for key, value in example.items()
    }


def _get_dataset_length(dataset) -> int:
    result = 0
    for _ in dataset:
        result += 1
    return result


class MyDatasets(NamedTuple):
    train_dataset: Any
    validation_dataset: Any
    test_dataset: Any


def _create_test_data() -> MyDatasets:
    # Generate small test dataset.
    train_samples_per_digit = 8
    validation_samples_per_digit = 3
    test_samples_per_digit = 2
    total_samples_per_digit = (
        train_samples_per_digit + validation_samples_per_digit + test_samples_per_digit
    )

    images_list: list[np.ndarray] = []
    labels_list: list[int] = []
    splits_list: list[str] = []
    for digit in range(10):
        # Encode digits as intensity and add some noise.
        images_list.append(
            float(digit)
            * np.ones(
                shape=(total_samples_per_digit, 1, 28, 28),
                dtype=np.float32,
            )
        )
        labels_list.extend([digit] * total_samples_per_digit)
        splits_list.extend(["train"] * train_samples_per_digit)
        splits_list.extend(["validation"] * validation_samples_per_digit)
        splits_list.extend(["test"] * test_samples_per_digit)

    images = np.concatenate(images_list) / 11.0
    labels = np.stack(labels_list)
    splits = np.stack(splits_list)

    np.random.seed(42)
    images = images + np.random.random_sample(images.shape) * 0.05

    # Shuffle all arrays in unison.
    indices = np.random.permutation(len(labels))
    images = np.asarray(images[indices, ...], dtype=np.float32)  # cast as float32
    labels = labels[indices]
    splits = splits[indices]

    with (
        tempfile.TemporaryDirectory() as tmp_dir,
        tempfile.NamedTemporaryFile(
            mode="wb", suffix=".npz", dir=tmp_dir
        ) as tmp_npz_file,
    ):
        npz_filename = tmp_npz_file.name
        np.savez_compressed(npz_filename, images=images, labels=labels, split=splits)

        dataset_cls = REGISTRY.get("npz_file", RegistryType.DATASET)
        assert dataset_cls is not None

        train_dataset = build_dataset(
            dataset_name="npz_file",
            tokenizer=None,
            split="train",
            use_torchdata=True,
            dataset_path=npz_filename,
        )
        assert _get_dataset_length(train_dataset) == 10 * train_samples_per_digit

        validation_dataset = build_dataset(
            dataset_name="npz_file",
            tokenizer=None,
            split="validation",
            use_torchdata=True,
            dataset_path=npz_filename,
        )
        assert (
            _get_dataset_length(validation_dataset) == 10 * validation_samples_per_digit
        )

        test_dataset = build_dataset(
            dataset_name="npz_file",
            tokenizer=None,
            split="test",
            use_torchdata=True,
            dataset_path=npz_filename,
        )
        assert _get_dataset_length(test_dataset) == 10 * test_samples_per_digit

        return MyDatasets(
            train_dataset=train_dataset,
            validation_dataset=validation_dataset,
            test_dataset=test_dataset,
        )


def _validate_prediction(model: torch.nn.Module, dataset, *, max_iters: int = 5):
    with torch.no_grad():
        model_device = next(model.parameters()).device
        batch_size = 2

        for idx, test_input in enumerate(DataLoader(dataset, batch_size=batch_size)):
            if idx >= max_iters:
                break
            inputs = _convert_example_to_model_input(test_input, device=model_device)

            # Without label
            outputs = model(images=inputs["images"])
            assert outputs.keys() == ({"logits"})
            assert isinstance(outputs["logits"], torch.Tensor)
            logits = outputs["logits"].cpu().numpy()
            assert logits.shape == (batch_size, 10)
            assert logits.dtype == np.float32

            # With label
            outputs = model(images=inputs["images"], labels=inputs["labels"])
            assert outputs.keys() == ({"logits", "loss"})
            assert isinstance(outputs["logits"], torch.Tensor)
            logits = outputs["logits"].cpu().numpy()
            assert logits.shape == (batch_size, 10)
            assert logits.dtype == np.float32
            loss = outputs["loss"].cpu().numpy()
            assert loss.shape == ()
            assert loss.dtype == np.float32


def test_basic_training_and_prediction():
    test_data = _create_test_data()

    # Create a model.
    model_params = ModelParams(
        model_name="CnnClassifier",
        load_pretrained_weights=False,
        model_kwargs={
            "image_width": 28,
            "image_height": 28,
            "in_channels": 1,
            "output_dim": 10,
        },
    )
    model = build_model(model_params)

    with (
        tempfile.TemporaryDirectory() as tmp_dir,
    ):
        training_params = TrainingParams(
            output_dir=str(Path(tmp_dir) / "cnn_classifier_test_output"),
            per_device_train_batch_size=32,
            per_device_eval_batch_size=8,
            eval_strategy="steps",
            eval_steps=50,
            max_steps=100,
            save_steps=0,
            logging_steps=50,
            optimizer="adam",
            learning_rate=2e-3,
        )

        trainer = Trainer(
            model=model,
            processing_class=None,  # No tokenizer! The custom model is non-textual
            args=training_params,
            train_dataset=test_data.train_dataset,
            dataloader_num_workers=2,
            dataloader_prefetch_factor=32,
        )

        trainer.train()

        _validate_prediction(trainer.model, test_data.test_dataset)



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module provides access to various judge configurations for the Oumi project.

The judges are used to evaluate the quality of AI-generated responses based on
different criteria such as helpfulness, honesty, and safety.
"""

from oumi.judges.base_judge import (
    BaseJudge,
    JudgeOutput,
    JudgeOutputField,
)
from oumi.judges.simple_judge import SimpleJudge

__all__ = [
    "BaseJudge",
    "SimpleJudge",
    "JudgeOutputField",
    "JudgeOutput",
]



#    Copyright 2023 Haotian Liu
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


from typing import List, Optional, Tuple, Union

import torch
import torch.nn as nn
from transformers import AutoConfig, AutoModelForCausalLM
from transformers.generation.utils import GenerateOutput
from transformers.modeling_outputs import CausalLMOutputWithPast

from oumi.models.experimental.cambrian.utils import IS_XLA_AVAILABLE

from ..cambrian_arch import CambrianMetaForCausalLM, CambrianMetaModel
from .phi3 import Phi3Config, Phi3ForCausalLM, Phi3Model


class CambrianConfig(Phi3Config):
    model_type = "cambrian_phi3"

    debug = "debug"


class CambrianPhi3Model(CambrianMetaModel, Phi3Model):
    config_class = CambrianConfig

    def __init__(self, config: Phi3Config):
        super().__init__(config)


class CambrianPhi3ForCausalLM(Phi3ForCausalLM, CambrianMetaForCausalLM):
    config_class = CambrianConfig

    def __init__(self, config):
        super(Phi3ForCausalLM, self).__init__(config)

        self.model = CambrianPhi3Model(config)
        self.vocab_size = config.vocab_size
        self.lm_head = nn.Linear(config.hidden_size, config.vocab_size, bias=False)

        # Initialize weights and apply final processing
        self.post_init()

    def get_model(self):
        return self.model

    def forward(
        self,
        input_ids: torch.LongTensor = None,
        attention_mask: Optional[torch.Tensor] = None,
        position_ids: Optional[torch.LongTensor] = None,
        past_key_values: Optional[List[torch.FloatTensor]] = None,
        inputs_embeds: Optional[torch.FloatTensor] = None,
        labels: Optional[torch.LongTensor] = None,
        use_cache: Optional[bool] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        images: Optional[torch.FloatTensor] = None,
        image_aux_attention_masks_list: Optional[List[torch.Tensor]] = None,
        image_sizes: Optional[List[List[int]]] = None,
        return_dict: Optional[bool] = None,
    ) -> Union[Tuple, CausalLMOutputWithPast]:
        if inputs_embeds is None:
            (
                input_ids,
                position_ids,
                attention_mask,
                past_key_values,
                inputs_embeds,
                labels,
                vision_tower_aux_feature_list,
                vision_tower_aux_attention_masks_list,
                final_vision_feature_size,
                global_context_feature,
            ) = self.prepare_inputs_labels_for_multimodal(
                input_ids,
                position_ids,
                attention_mask,
                past_key_values,
                labels,
                images,
                image_aux_attention_masks_list,
                image_sizes,
            )

        if IS_XLA_AVAILABLE:
            # Very Important for TorchXLA
            # self.model.gradient_checkpointing = False

            from torch_xla.utils.checkpoint import checkpoint

            self.model._gradient_checkpointing_func = checkpoint

        output_attentions = (
            output_attentions
            if output_attentions is not None
            else self.config.output_attentions
        )
        output_hidden_states = (
            output_hidden_states
            if output_hidden_states is not None
            else self.config.output_hidden_states
        )
        return_dict = (
            return_dict if return_dict is not None else self.config.use_return_dict
        )

        if IS_XLA_AVAILABLE:
            output = super().forward(
                input_ids=input_ids,
                attention_mask=attention_mask,
                position_ids=position_ids,
                past_key_values=past_key_values,
                inputs_embeds=inputs_embeds,
                labels=labels,
                use_cache=use_cache,
                output_attentions=output_attentions,
                output_hidden_states=output_hidden_states,
                return_dict=return_dict,
                vision_tower_aux_feature_list=vision_tower_aux_feature_list,
                vision_tower_aux_attention_masks_list=vision_tower_aux_attention_masks_list,
                final_vision_feature_size=final_vision_feature_size,
                global_context_feature=global_context_feature,
            )

        else:
            output = super().forward(
                input_ids=input_ids,
                attention_mask=attention_mask,
                position_ids=position_ids,
                past_key_values=past_key_values,
                inputs_embeds=inputs_embeds,
                labels=labels,
                use_cache=use_cache,
                output_attentions=output_attentions,
                output_hidden_states=output_hidden_states,
                return_dict=return_dict,
                vision_tower_aux_feature_list=vision_tower_aux_feature_list
                if inputs_embeds is None
                else self.vision_tower_aux_feature_list,
                vision_tower_aux_attention_masks_list=vision_tower_aux_attention_masks_list
                if inputs_embeds is None
                else self.vision_tower_aux_attention_masks_list,
                final_vision_feature_size=final_vision_feature_size
                if inputs_embeds is None
                else self.final_vision_feature_size,
                global_context_feature=global_context_feature
                if inputs_embeds is None
                else self.global_context_feature,
            )

        return output

    @torch.no_grad()
    def generate(
        self,
        inputs: Optional[torch.Tensor] = None,
        images: Optional[torch.Tensor] = None,
        image_sizes: Optional[torch.Tensor] = None,
        **kwargs,
    ) -> Union[GenerateOutput, torch.LongTensor]:
        position_ids = kwargs.pop("position_ids", None)
        attention_mask = kwargs.pop("attention_mask", None)
        if "inputs_embeds" in kwargs:
            raise NotImplementedError("`inputs_embeds` is not supported")

        if images is not None:
            (
                inputs,
                position_ids,
                attention_mask,
                _,
                inputs_embeds,
                _,
                vision_tower_aux_feature_list,
                vision_tower_aux_attention_masks_list,
                final_vision_feature_size,
                global_context_feature,
            ) = self.prepare_inputs_labels_for_multimodal(
                inputs,
                position_ids,
                attention_mask,
                None,
                None,
                images,
                image_sizes=image_sizes,
            )
            self.vision_tower_aux_feature_list = vision_tower_aux_feature_list
            self.vision_tower_aux_attention_masks_list = (
                vision_tower_aux_attention_masks_list
            )
            self.final_vision_feature_size = final_vision_feature_size
            self.global_context_feature = global_context_feature
        else:
            inputs_embeds = self.get_model().embed_tokens(inputs)

        return super().generate(
            position_ids=position_ids,
            attention_mask=attention_mask,
            inputs_embeds=inputs_embeds,
            **kwargs,
        )

    def prepare_inputs_for_generation(
        self, input_ids, past_key_values=None, inputs_embeds=None, **kwargs
    ):
        images = kwargs.pop("images", None)
        image_sizes = kwargs.pop("image_sizes", None)
        inputs = super().prepare_inputs_for_generation(
            input_ids,
            past_key_values=past_key_values,
            inputs_embeds=inputs_embeds,
            **kwargs,
        )
        if images is not None:
            inputs["images"] = images
        if image_sizes is not None:
            inputs["image_sizes"] = image_sizes
        return inputs


AutoConfig.register("cambrian_phi3", CambrianConfig)
AutoModelForCausalLM.register(CambrianConfig, CambrianPhi3ForCausalLM)

# Used in the following published models:
# nyu-visionx/cambrian-phi3-3b
